import importlib

import pytest

from careamics.dataset_ng.image_stack import CziImageStack


def test_import_czi_image_stack_failure():
    """Test that in the absence of pylibCZIrw, the CziImageStack import fails
    with an import error."""
    if importlib.util.find_spec("pylibCZIrw") is None:
        with pytest.raises(ImportError):
            CziImageStack(data_path="")



"""A protocol for patch filtering."""

from typing import Protocol

from careamics.dataset_ng.patching_strategies import PatchSpecs


class CoordinateFilterProtocol(Protocol):
    """
    An interface for implementing coordinate filtering strategies.
    """

    def filter_out(self, patch: PatchSpecs) -> bool:
        """
        Determine whether to filter out a given patch based on its coordinates.

        Parameters
        ----------
        patch : PatchSpecs
            The patch coordinates to evaluate.

        Returns
        -------
        bool
            True if the patch should be filtered out (excluded), False otherwise.
        """
        ...



import pytest

from careamics.config.validators import (
    patch_size_ge_than_8_power_of_2,
)


@pytest.mark.parametrize(
    "patch_size, error",
    [
        ((2, 8, 8), True),
        ((10,), True),
        ((8, 10, 16), True),
        ((8, 13), True),
        ((8, 16, 4), True),
        ((8,), False),
        ((8, 8), False),
        ((8, 64, 64), False),
    ],
)
def test_patch_size(patch_size, error):
    """Test if patch size is valid."""
    if error:
        with pytest.raises(ValueError):
            patch_size_ge_than_8_power_of_2(patch_size)
    else:
        patch_size_ge_than_8_power_of_2(patch_size)



"""Module containing functions to convert prediction outputs to desired form."""

from typing import Any, Literal, Union, overload

import numpy as np
from numpy.typing import NDArray

from ..config.data.tile_information import TileInformation
from .stitch_prediction import stitch_prediction, stitch_prediction_vae


def convert_outputs(predictions: list[Any], tiled: bool) -> list[NDArray]:
    """
    Convert the Lightning trainer outputs to the desired form.

    This method allows stitching back together tiled predictions.

    Parameters
    ----------
    predictions : list
        Predictions that are output from `Trainer.predict`.
    tiled : bool
        Whether the predictions are tiled.

    Returns
    -------
    list of numpy.ndarray or numpy.ndarray
        list of arrays with the axes SC(Z)YX. If there is only 1 output it will not
        be in a list.
    """
    if len(predictions) == 0:
        return predictions

    # this layout is to stop mypy complaining
    if tiled:
        predictions_comb = combine_batches(predictions, tiled)
        predictions_output = stitch_prediction(*predictions_comb)
    else:
        predictions_output = combine_batches(predictions, tiled)

    return predictions_output


def convert_outputs_pn2v(
    predictions: list[Any], tiled: bool
) -> tuple[list[NDArray], list[NDArray]]:
    """
    Convert the Lightning trainer outputs to the desired form.

    This method allows stitching back together tiled predictions.

    Parameters
    ----------
    predictions : list
        Predictions that are output from `Trainer.predict`. Length of list the total
        number of tiles divided by the batch size. Each element consists of a tuple of
        ((prediction, mse), tile_info_list). 1st dimension of each tensor is the bs.
        Length of tile info list is the batch size.

    tiled : bool
        Whether the predictions are tiled.

    Returns
    -------
    tuple[list[NDArray], list[NDArray]]
        Tuple of (predictions, mmse) where each is a list of arrays with axes SC(Z)YX.
    """
    if len(predictions) == 0:
        return [], []
    # TODO test with multi_channel predictions
    if tiled:
        # Separate predictions and mmse, keeping tile info for each
        pred_with_tiles = [
            (pred, tile_info_list) for (pred, _), tile_info_list in predictions
        ]
        mse_with_tiles = [
            (mse, tile_info_list) for (_, mse), tile_info_list in predictions
        ]

        # Process predictions
        pred_comb = combine_batches(pred_with_tiles, tiled)
        predictions_output = stitch_prediction(*pred_comb)

        # Process mmse
        mse_comb = combine_batches(mse_with_tiles, tiled)
        mse_output = stitch_prediction(*mse_comb)

        return predictions_output, mse_output
    else:
        # Separate predictions and mmse for non-tiled case
        pred_only_tuple, mse_only_tuple = zip(*predictions, strict=False)
        pred_only_list: list[NDArray] = list(pred_only_tuple)
        mse_only_list: list[NDArray] = list(mse_only_tuple)

        predictions_output = combine_batches(pred_only_list, tiled=False)
        mse_output = combine_batches(mse_only_list, tiled=False)

        return predictions_output, mse_output


def convert_outputs_microsplit(
    predictions: list[tuple[NDArray, NDArray]], dataset
) -> tuple[NDArray, NDArray]:
    """
    Convert microsplit Lightning trainer outputs using eval_utils stitching functions.

    This function processes microsplit predictions that return
    (tile_prediction, tile_std) tuples and stitches them back together using the same
    logic as get_single_file_mmse.

    Parameters
    ----------
    predictions : list of tuple[NDArray, NDArray]
        Predictions from Lightning trainer for microsplit. Each element is a tuple of
        (tile_prediction, tile_std) where both are numpy arrays from predict_step.
    dataset : Dataset
        The dataset object used for prediction, needed for stitching function selection
        and stitching process.

    Returns
    -------
    tuple[NDArray, NDArray]
        A tuple of (stitched_predictions, stitched_stds) representing the full
        stitched predictions and standard deviations.
    """
    if len(predictions) == 0:
        raise ValueError("No predictions provided")

    # Separate predictions and stds from the list of tuples
    tile_predictions = [pred for pred, _ in predictions]
    tile_stds = [std for _, std in predictions]

    # Concatenate all tiles exactly like get_single_file_mmse
    tiles_arr = np.concatenate(tile_predictions, axis=0)
    tile_stds_arr = np.concatenate(tile_stds, axis=0)

    # Apply stitching using stitch_predictions_new
    stitched_predictions = stitch_prediction_vae(tiles_arr, dataset)
    stitched_stds = stitch_prediction_vae(tile_stds_arr, dataset)

    return stitched_predictions, stitched_stds


# for mypy
@overload
def combine_batches(  # numpydoc ignore=GL08
    predictions: list[Any], tiled: Literal[True]
) -> tuple[list[NDArray], list[TileInformation]]: ...


# for mypy
@overload
def combine_batches(  # numpydoc ignore=GL08
    predictions: list[Any], tiled: Literal[False]
) -> list[NDArray]: ...


# for mypy
@overload
def combine_batches(  # numpydoc ignore=GL08
    predictions: list[Any], tiled: Union[bool, Literal[True], Literal[False]]
) -> Union[list[NDArray], tuple[list[NDArray], list[TileInformation]]]: ...


def combine_batches(
    predictions: list[Any], tiled: bool
) -> Union[list[NDArray], tuple[list[NDArray], list[TileInformation]]]:
    """
    If predictions are in batches, they will be combined.

    # TODO improve description!

    Parameters
    ----------
    predictions : list
        Predictions that are output from `Trainer.predict`.
    tiled : bool
        Whether the predictions are tiled.

    Returns
    -------
    (list of numpy.ndarray) or tuple of (list of numpy.ndarray, list of TileInformation)
        Combined batches.
    """
    if tiled:
        return _combine_tiled_batches(predictions)
    else:
        return _combine_array_batches(predictions)


def _combine_tiled_batches(
    predictions: list[tuple[NDArray, list[TileInformation]]],
) -> tuple[list[NDArray], list[TileInformation]]:
    """
    Combine batches from tiled output.

    Parameters
    ----------
    predictions : list of (numpy.ndarray, list of TileInformation)
        Predictions that are output from `Trainer.predict`. For tiled batches, this is
        a list of tuples. The first element of the tuples is the prediction output of
        tiles with dimension (B, C, (Z), Y, X), where B is batch size. The second
        element of the tuples is a list of TileInformation objects of length B.

    Returns
    -------
    tuple of (list of numpy.ndarray, list of TileInformation)
        Combined batches.
    """
    # turn list of lists into single list
    tile_infos = [
        tile_info for *_, tile_info_list in predictions for tile_info in tile_info_list
    ]
    prediction_tiles: list[NDArray] = _combine_array_batches(
        [preds for preds, *_ in predictions]
    )

    return prediction_tiles, tile_infos


def _combine_array_batches(predictions: list[NDArray]) -> list[NDArray]:
    """
    Combine batches of arrays.

    Parameters
    ----------
    predictions : list
        Prediction arrays that are output from `Trainer.predict`. A list of arrays that
        have dimensions (B, C, (Z), Y, X), where B is batch size.

    Returns
    -------
    list of numpy.ndarray
        A list of arrays with dimensions (1, C, (Z), Y, X).
    """
    prediction_concat: NDArray = np.concatenate(predictions, axis=0)
    prediction_split = np.split(prediction_concat, prediction_concat.shape[0], axis=0)
    return prediction_split



"""PyTorch lightning utilities."""

from pathlib import Path
from typing import Union


def read_csv_logger(experiment_name: str, log_folder: Union[str, Path]) -> dict:
    """Return the loss curves from the csv logs.

    Parameters
    ----------
    experiment_name : str
        Name of the experiment.
    log_folder : Path or str
        Path to the folder containing the csv logs.

    Returns
    -------
    dict
        Dictionary containing the loss curves, with keys "train_epoch", "val_epoch",
        "train_loss" and "val_loss".
    """
    path = Path(log_folder) / experiment_name

    # find the most recent of version_* folders
    versions = [int(v.name.split("_")[-1]) for v in path.iterdir() if v.is_dir()]
    version = max(versions)

    path_log = path / f"version_{version}" / "metrics.csv"

    with open(path_log) as f:
        lines = f.readlines()

        header = lines[0].strip().split(",")
        metrics: dict[str, list] = {value: [] for value in header}
        print(metrics)

        for single_line in lines[1:]:
            values = single_line.strip().split(",")

            for k, v in zip(header, values, strict=False):
                metrics[k].append(v)

    # train and val are not logged on the same row and can have different lengths
    train_epoch = [
        int(metrics["epoch"][i])
        for i in range(len(metrics["epoch"]))
        if metrics["train_loss_epoch"][i] != ""
    ]
    val_epoch = [
        int(metrics["epoch"][i])
        for i in range(len(metrics["epoch"]))
        if metrics["val_loss"][i] != ""
    ]
    train_losses = [
        float(metrics["train_loss_epoch"][i])
        for i in range(len(metrics["train_loss_epoch"]))
        if metrics["train_loss_epoch"][i] != ""
    ]
    val_losses = [
        float(metrics["val_loss"][i])
        for i in range(len(metrics["val_loss"]))
        if metrics["val_loss"][i] != ""
    ]

    return {
        "train_epoch": train_epoch,
        "val_epoch": val_epoch,
        "train_loss": train_losses,
        "val_loss": val_losses,
    }



import numpy as np
import pytest

from careamics.dataset_ng.image_stack.image_utils.image_stack_utils import (
    channel_slice,
    pad_patch,
    reshape_array_shape,
)

data = np.array(
    [
        [
            [1, 2, 3],
            [4, 5, 6],
            [7, 8, 9],
        ]
    ]
)


@pytest.mark.parametrize(
    "data_source",
    [data.copy()],
)
@pytest.mark.parametrize(
    "coords, patch_size, expected_patch",
    [
        ((-1, 0), (2, 2), np.array([[[0, 0], [1, 2]]])),
        ((1, 1), (3, 3), np.array([[[5, 6, 0], [8, 9, 0], [0, 0, 0]]])),
        ((0, 0), (3, 3), data.copy()),
        ((-4, 5), (2, 2), np.zeros((1, 2, 2), dtype=data.dtype)),
        (
            (-1, -1),
            (3, 5),
            np.array([[[0, 0, 0, 0, 0], [0, 1, 2, 3, 0], [0, 4, 5, 6, 0]]]),
        ),
    ],
)
def test_pad_patch(data_source, coords, patch_size, expected_patch):
    patch = data_source[
        (
            slice(None, None, None),  # channel axis
            *[
                slice(np.clip(c, 0, s), np.clip(c + ps, 0, s), None)
                for c, ps, s in zip(
                    coords, patch_size, data_source.shape[1:], strict=False
                )
            ],
        )
    ]
    patch_padded = pad_patch(coords, patch_size, (1, *data_source.shape), patch)
    np.testing.assert_equal(patch_padded, expected_patch)


def test_channel_slice():
    """Test channel_slice utility function."""
    shape = (5, 32, 32)
    array = np.arange(np.prod(shape)).reshape(shape)

    # Test single channel
    single_channel = array[channel_slice([3])]
    assert single_channel.shape[0] == 1
    np.testing.assert_array_equal(single_channel, array[3:4])

    # Test multiple channels
    multiple_channels = array[channel_slice([0, 3, 4])]
    assert multiple_channels.shape[0] == 3
    np.testing.assert_array_equal(multiple_channels, array[[0, 3, 4]])

    # Test all channels
    all_channels = array[channel_slice(None)]  # only one ellipsis allowed
    assert all_channels.shape[0] == shape[0]
    np.testing.assert_array_equal(all_channels, array)


def test_error_empty_channel_slice():
    """Test that channel_slice raises an error for empty channel list."""
    with pytest.raises(ValueError):
        _ = channel_slice([])


@pytest.mark.parametrize(
    "axes, shape, expected_shape, add_singleton",
    [
        # axes order (no T)
        ("YZSXC", (1, 2, 3, 4, 5), (3, 5, 2, 1, 4), False),
        ("YZSXC", (1, 2, 3, 4, 5), (3, 5, 2, 1, 4), True),
        ("XYZS", (1, 2, 3, 4), (4, 3, 2, 1), False),
        ("XYZS", (1, 2, 3, 4), (4, 1, 3, 2, 1), True),
        # multiplex S and T
        ("STCYX", (2, 3, 4, 5, 6), (6, 4, 5, 6), False),
        ("STCYX", (2, 3, 4, 5, 6), (6, 4, 5, 6), True),
        ("CTYSX", (2, 3, 4, 5, 6), (15, 2, 4, 6), False),
        ("CTYSX", (2, 3, 4, 5, 6), (15, 2, 4, 6), True),
        # T but no S
        ("TYXC", (2, 3, 4, 5), (2, 5, 3, 4), False),
        ("TYXC", (2, 3, 4, 5), (2, 5, 3, 4), True),
        # missing S, C, or T (singleton added or not)
        ("YX", (32, 32), (1, 1, 32, 32), True),
        ("YX", (32, 32), (32, 32), False),
        ("YXC", (32, 32, 3), (1, 3, 32, 32), True),
        ("YXC", (32, 32, 3), (3, 32, 32), False),
        ("SYX", (2, 32, 32), (2, 1, 32, 32), True),
        ("SYX", (2, 32, 32), (2, 32, 32), False),
        # random letters
        ("SABYX", (2, 3, 4, 5, 6), (2, 1, 5, 6), True),
    ],
)
def test_reshape_array_shape(axes, shape, expected_shape, add_singleton):
    """Test reshape_array_shape utility function."""
    reshaped_shape = reshape_array_shape(axes, shape, add_singleton)
    assert reshaped_shape == expected_shape



import numpy as np
import torch
from numpy.typing import NDArray

from .normalization_protocol import NormalizationProtocol


def _reshape_stats(stats: list[float], ndim: int, channel_axis: int = 0) -> NDArray:
    """Reshape stats to match the number of dimensions of the input image.

    Parameters
    ----------
    stats : list of float
        List of per-channel statistic values.
    ndim : int
        Number of dimensions in the image.
    channel_axis : int, optional
        Position of the channel axis in the image.

    Returns
    -------
    NDArray
        Stats array reshaped, with the channel dimension
        at the specified axis and singleton dimensions elsewhere.
    """
    shape = [1] * ndim
    shape[channel_axis] = len(stats)
    return np.array(stats, dtype=np.float32).reshape(shape)


class MeanStdNormalization(NormalizationProtocol):
    """
    Normalize an image or image patch.

    Normalization is a zero mean and unit variance. This transform expects C(Z)YX
    dimensions.

    Note that an epsilon value of 1e-6 is added to the standard deviation to avoid
    division by zero and that it returns a float32 image.

    Parameters
    ----------
    input_means : list of float
        Mean value per channel.
    input_stds : list of float
        Standard deviation value per channel.
    target_means : list of float, optional
        Target mean value per channel, by default None.
    target_stds : list of float, optional
        Target standard deviation value per channel, by default None.

    Attributes
    ----------
    input_means : list of float
        Mean value per channel.
    input_stds : list of float
        Standard deviation value per channel.
    target_means : list of float, optional
        Target mean value per channel, by default None.
    target_stds : list of float, optional
        Target standard deviation value per channel, by default None.
    """

    def __init__(
        self,
        input_means: list[float],
        input_stds: list[float],
        target_means: list[float] | None = None,
        target_stds: list[float] | None = None,
    ):
        """Constructor.

        Parameters
        ----------
        input_means : list of float
            Mean value per channel.
        input_stds : list of float
            Standard deviation value per channel.
        target_means : list of float, optional
            Target mean value per channel, by default None.
        target_stds : list of float, optional
            Target standard deviation value per channel, by default None.
        """
        self.input_means = input_means
        self.input_stds = input_stds
        self.target_means = target_means
        self.target_stds = target_stds

        self.eps = 1e-6

    def __call__(
        self,
        patch: np.ndarray,
        target: NDArray | None = None,
    ) -> tuple[NDArray, NDArray | None]:
        """Apply the transform to the source patch and the target (optional).

        Parameters
        ----------
        patch : NDArray
            Patch, 2D or 3D, shape C(Z)YX.
        target : NDArray, optional
            Target for the patch, by default None.

        Returns
        -------
        tuple of NDArray
            Transformed patch and target, the target can be returned as `None`.
        """
        if len(self.input_means) != patch.shape[0]:
            raise ValueError(
                f"Number of means (got a list of size {len(self.input_means)}) and "
                f"number of channels (got shape {patch.shape} for C(Z)YX) do not match."
            )

        patch = patch.astype(np.float32)
        if target is not None:
            target = target.astype(np.float32)

        # reshape mean and std and apply the normalization to the patch
        means = _reshape_stats(self.input_means, patch.ndim, channel_axis=0)
        stds = _reshape_stats(self.input_stds, patch.ndim, channel_axis=0)
        norm_patch = self._apply_normalization(patch, means, stds)

        # same for the target patch
        if target is None:
            norm_target = None
        else:
            if not self.target_means or not self.target_stds:
                raise ValueError(
                    "Target means and standard deviations must be provided "
                    "if target is not None."
                )
            target_means = _reshape_stats(
                self.target_means, target.ndim, channel_axis=0
            )
            target_stds = _reshape_stats(self.target_stds, target.ndim, channel_axis=0)
            norm_target = self._apply_normalization(target, target_means, target_stds)

        return norm_patch, norm_target

    def _apply_normalization(
        self, patch: NDArray, mean: NDArray, std: NDArray
    ) -> NDArray:
        """
        Apply the transform to the image.

        Parameters
        ----------
        patch : NDArray
            Image patch, 2D or 3D, shape C(Z)YX.
        mean : NDArray
            Mean values.
        std : NDArray
            Standard deviations.

        Returns
        -------
        NDArray
            Normalized image patch.
        """
        return (patch - mean) / (std + self.eps)

    def _apply_denormalization(
        self, array: torch.Tensor, mean: NDArray, std: NDArray
    ) -> torch.Tensor:
        """
        Apply the transform to the image.

        Parameters
        ----------
        array : torch.Tensor
            Image patch, 2D or 3D, shape BC(Z)YX.
        mean : NDArray
            Mean values.
        std : NDArray
            Standard deviations.

        Returns
        -------
        torch.Tensor
            Denormalized image array.
        """
        mean_tensor = torch.from_numpy(mean).to(array.device)
        std_tensor = torch.from_numpy(std).to(array.device)
        return array * (std_tensor + self.eps) + mean_tensor

    def denormalize(self, patch: torch.Tensor) -> torch.Tensor:
        """Reverse the normalization operation for a batch of patches.

        Parameters
        ----------
        patch : torch.Tensor
            Patch, 2D or 3D, shape BC(Z)YX.

        Returns
        -------
        torch.Tensor
            Transformed array.
        """
        if len(self.input_means) != patch.shape[1]:
            raise ValueError(
                f"Number of means (got a list of size {len(self.input_means)}) and "
                f"number of channels (got shape {patch.shape} for BC(Z)YX) do not "
                f"match."
            )

        patch = patch.to(dtype=torch.float32)

        means = _reshape_stats(self.input_means, patch.ndim, channel_axis=1)
        stds = _reshape_stats(self.input_stds, patch.ndim, channel_axis=1)

        denorm_array = self._apply_denormalization(patch, means, stds)

        return denorm_array



"""Package to house various prediction utilies."""

__all__ = [
    "convert_outputs",
    "convert_outputs_microsplit",
    "convert_outputs_pn2v",
    "stitch_prediction",
    "stitch_prediction_single",
    "stitch_prediction_vae",
]

from .prediction_outputs import (
    convert_outputs,
    convert_outputs_microsplit,
    convert_outputs_pn2v,
)
from .stitch_prediction import (
    stitch_prediction,
    stitch_prediction_single,
    stitch_prediction_vae,
)



"""Supported normalization strategies for CAREamics."""

from careamics.utils import BaseEnum


class SupportedNormalization(str, BaseEnum):
    """Normalization strategies supported by Careamics."""

    MEAN_STD = "mean_std"
    """Mean and std normalization strategy."""

    QUANTILE = "quantile"
    """Quantile normalization strategy."""

    MINMAX = "minmax"
    """Min-max normalization strategy."""

    NONE = "none"
    """No normalization strategy."""



"""Tiled prediction stitching utilities."""

import builtins
from collections import defaultdict
from typing import Literal

import numpy as np
from numpy.typing import NDArray

from careamics.dataset_ng.dataset import ImageRegionData
from careamics.dataset_ng.patching_strategies import TileSpecs


def group_tiles_by_key(
    tiles: list[ImageRegionData], key: Literal["data_idx", "sample_idx"]
) -> dict[int, list[ImageRegionData]]:
    """
    Sort tiles by key.

    Parameters
    ----------
    tiles : list of ImageRegionData
        List of tiles to sort.
    key : {'data_idx', 'sample_idx'}
        Key to group tiles by.

    Returns
    -------
    {int: list of ImageRegionData}
        Dictionary mapping data indices to lists of tiles.
    """
    sorted_tiles: dict[int, list[ImageRegionData]] = defaultdict(list)
    for tile in tiles:
        key_value = tile.region_spec[key]
        sorted_tiles[key_value].append(tile)
    return sorted_tiles


def stitch_prediction(
    tiles: list[ImageRegionData],
) -> tuple[list[NDArray], list[str]]:
    """
    Stitch tiles back together to form full images.

    Tiles are of dimensions SC(Z)YX, where C is the number of channels and can be a
    singleton dimension.

    Parameters
    ----------
    tiles : list of ImageRegionData
        Cropped tiles and their respective stitching coordinates. Can contain tiles
        from multiple images.

    Returns
    -------
    list of numpy.ndarray
        Full images, may be a single image.
    list of str
        List of sources, one per output.
    """
    # sort tiles by data index
    grouped_tiles: dict[int, list[ImageRegionData]] = group_tiles_by_key(
        tiles, key="data_idx"
    )

    # stitch each image separately
    image_predictions: list[NDArray] = []
    image_sources: list[str] = []
    for data_idx in sorted(grouped_tiles.keys()):
        image_predictions.append(stitch_single_prediction(grouped_tiles[data_idx]))
        image_sources.append(grouped_tiles[data_idx][0].source)

    return image_predictions, image_sources


def stitch_single_prediction(
    tiles: list[ImageRegionData],
) -> NDArray:
    """
    Stitch tiles back together to form a full image.

    Tiles are of dimensions C(Z)YX, where C is the number of channels and can be a
    singleton dimension.

    Parameters
    ----------
    tiles : list of ImageRegionData
        Cropped tiles and their respective stitching coordinates.

    Returns
    -------
    numpy.ndarray
        Full image, with dimensions SC(Z)YX.
    """
    data_shape = tiles[0].data_shape
    predicted_image = np.zeros(data_shape, dtype=np.float32)

    if "S" in tiles[0].axes:
        tiles_by_sample = group_tiles_by_key(tiles, key="sample_idx")
        for sample_idx in tiles_by_sample.keys():
            sample_tiles = tiles_by_sample[sample_idx]
            stitched_sample = stitch_single_sample(sample_tiles)

            # compute sample slice
            sample_slice = slice(
                sample_idx,
                sample_idx + 1,
            )

            # insert stitched sample into predicted image
            predicted_image[sample_slice] = stitched_sample.astype(np.float32)
    else:
        # stitch as a single sample
        # predicted_image has singleton sample dimension
        predicted_image[0] = stitch_single_sample(tiles)

    return predicted_image


def stitch_single_sample(
    tiles: list[ImageRegionData],
) -> NDArray:
    """
    Stitch tiles back together to form a full sample.

    Tiles are of dimensions C(Z)YX, where C is the number of channels and can be a
    singleton dimension.

    Parameters
    ----------
    tiles : list of ImageRegionData
        Cropped tiles and their respective stitching coordinates.

    Returns
    -------
    numpy.ndarray
        Full sample, with dimensions C(Z)YX.
    """
    data_shape = tiles[0].data_shape  # SC(Z)YX
    predicted_sample = np.zeros(data_shape[1:], dtype=np.float32)

    for tile in tiles:
        # compute crop coordinates and stitiching coordinates
        tile_spec: TileSpecs = tile.region_spec  # type: ignore
        crop_coords = tile_spec["crop_coords"]
        crop_size = tile_spec["crop_size"]
        stitch_coords = tile_spec["stitch_coords"]

        crop_slices: tuple[builtins.ellipsis | slice, ...] = (
            ...,
            *[
                slice(start, start + length)
                for start, length in zip(crop_coords, crop_size, strict=True)
            ],
        )

        stitch_slices: tuple[builtins.ellipsis | slice, ...] = (
            ...,
            *[
                slice(start, start + length)
                for start, length in zip(stitch_coords, crop_size, strict=True)
            ],
        )

        # crop predited tile according to overlap coordinates
        cropped_tile = tile.data[crop_slices]

        # insert cropped tile into predicted image
        predicted_sample[stitch_slices] = cropped_tile.astype(np.float32)

    return predicted_sample



"""Module containing file path utilities for `WriteStrategy` to use."""

from pathlib import Path


def create_write_file_path(
    dirpath: Path, file_path: Path, write_extension: str, postfix: str = ""
) -> Path:
    """
    Create the file name for the output file.

    Takes the original file path, changes the directory to `dirpath` and changes
    the extension to `write_extension`.

    Parameters
    ----------
    dirpath : pathlib.Path
        The output directory to write file to.
    file_path : pathlib.Path
        The original file path.
    write_extension : str
        The extension that output files should have.
    postfix : str, default=""
        Appends to filename before extension.

    Returns
    -------
    Path
        The output file path.
    """
    file_path = Path(file_path)  # as a guard against str input
    file_name = f"{file_path.stem}{postfix}"
    file_path = dirpath / Path(file_name).with_suffix(write_extension)
    return file_path



from collections.abc import Sequence
from types import EllipsisType
from typing import TypeVar

import numpy as np
from numpy.typing import NDArray

T = TypeVar("T", bound=np.generic)


def channel_slice(
    channels: Sequence[int] | None,
) -> EllipsisType | Sequence[int]:
    """Create a slice or sequence for indexing channels while preserving dimensions.

    Parameters
    ----------
    channels : Sequence[int] | None
        The channel indices to select, or None to select all channels.

    Returns
    -------
    EllipsisType | Sequence[int]
        An indexing object that can be used to index the channel dimension while
        preserving it.
    """
    if channels is None:
        return ...

    if len(channels) == 0:
        raise ValueError("Channel index sequence cannot be empty.")

    return channels


# TODO: add tests
# TODO: move to dataset_utils, better name?
def reshape_array_shape(
    original_axes: str, shape: Sequence[int], add_singleton: bool = True
) -> tuple[int, ...]:
    """Find resulting shape if reshaping array to SC(Z)YX.

    If `T` is present in the original axes, its size is multiplied into `S`, as both
    axes are multiplexed.

    Setting `add_singleton` to `False` will only include axes that are present in
    `original_axes` in the output shape.

    Parameters
    ----------
    original_axes : str
        The axes of the original array, e.g. "TCZYX", "SCYX", etc.
    shape : Sequence[int]
        The shape of the original array.
    add_singleton : bool, default=True
        Whether to add singleton dimensions for missing axes. When `False`, only axes
        present in `original_axes` will be included in the output shape. When `True`,
        missing mandatory axes (`S` and `C`) will be added as singleton dimensions.
    """
    target_axes = "SCZYX"
    target_shape = []
    for d in target_axes:
        if d in original_axes:
            idx = original_axes.index(d)
            target_shape.append(shape[idx])
        elif d != "Z":
            if add_singleton:
                target_shape.append(1)

    if "T" in original_axes:
        idx = original_axes.index("T")
        if "S" in original_axes or add_singleton:
            target_shape[0] = target_shape[0] * shape[idx]
        else:
            target_shape.insert(0, shape[idx])

    return tuple(target_shape)


def pad_patch(
    coords: Sequence[int],
    patch_size: Sequence[int],
    data_shape: Sequence[int],
    patch_data: NDArray[T],
) -> NDArray[T]:
    """
    Pad patch data with zeros where it is outside the bounds of it's source image.

    This ensures the patch data is contained in an array with the expected patch size.

    If `coords` are negative, the start of the patch will be padded with zeros up until
    where the start of the image would be, and this is where the patch data starts.

    If the `coords + patch_size` are greater than the bounds of the image then the
    end of the patch will be filled with zeros.

    Parameters
    ----------
    coords : Sequence[int]
        The coordinates that describe where the patch starts in the spatial dimension of
        the image
    patch_size : Sequence[int]
        The size of the patch in the spatial dimensions.
    data_shape : Sequence[int]
        The shape of the image the patch originates from, must be in the format SC(Z)YX.
    patch_data : NDArray[T]
        The patch data to be padded.

    Returns
    -------
    NDArray[T]
        The resulting padded patch.
    """
    coords_ = np.array(coords)
    patch = np.zeros((patch_data.shape[0], *patch_size), dtype=patch_data.dtype)
    # data start will be zero unless coords are negative
    data_start = np.clip(coords_, 0, None) - coords_
    data_end = data_start + np.array(patch_data.shape[1:])
    patch[
        (
            slice(None, None, None),  # channel slice
            *tuple(slice(s, t) for s, t in zip(data_start, data_end, strict=False)),
        )
    ] = patch_data
    return patch



from collections.abc import Sequence
from pathlib import Path
from typing import Any, Generic, Literal, NamedTuple, Union

import numpy as np
from numpy.typing import NDArray
from torch.utils.data import Dataset

from careamics.config.data.ng_data_config import Mode, NGDataConfig, WholePatchingConfig
from careamics.transforms import Compose

from .image_stack import GenericImageStack, ZarrImageStack
from .normalization import create_normalization
from .normalization.statistics import resolve_normalization_config
from .patch_extractor import PatchExtractor
from .patch_filter import create_coord_filter, create_patch_filter
from .patching_strategies import (
    PatchSpecs,
    RegionSpecs,
    create_patching_strategy,
)


class ImageRegionData(NamedTuple, Generic[RegionSpecs]):
    """
    Data structure for arrays produced by the dataset and propagated through models.

    An ImageRegionData may be a patch during training/validation, a tile during
    prediction with tiling, or a whole image during prediction without tiling.

    `data_shape` may not correspond to the shape of the original data if a subset
    of the channels has been requested, in which case the channel dimension may
    be smaller than that of the original data and only correspond to the requested
    number of channels.

    ImageRegionData may be collated in batches during training by the DataLoader. In
    that case:
    - data: arrays are collated into NDArray of shape (B,C,Z,Y,X)
    - source: list of str, length B
    - data_shape: list of tuples of int, each tuple being of length B and representing
        the shape of the original images in the corresponding dimension
    - dtype: list of str, length B
    - axes: list of str, length B
    - region_spec: dict of {str: sequence}, each sequence being of length B
    - additional_metadata: list of dict

    Description of the fields is given for the uncollated case (non-batched).
    """

    data: NDArray
    """Patch, tile or image in C(Z)YX format."""

    source: Union[str, Literal["array"]]
    """Source of the data, e.g. file path, zarr URI, or "array" for in-memory arrays."""

    data_shape: Sequence[int]
    """Shape of the original image in (SCZ)YX format and order. If channels are
    subsetted, the channel dimension corresponds to the number of requested channels."""

    dtype: str  # dtype should be str for collate
    """Data type of the original image as a string."""

    axes: str
    """Axes of the original data array, in format SCZYX."""

    region_spec: RegionSpecs  # PatchSpecs or subclasses, e.g. TileSpecs
    """Specifications of the region within the original image from where `data` is
    extracted. Of type PatchSpecs during training/validation and prediction without
    tiling, and TileSpecs during prediction with tiling.
    """

    additional_metadata: dict[str, Any]
    """Additional metadata to be stored with the image region. Currently used to store
    chunk and shard information for zarr image stacks."""


InputType = Union[Sequence[NDArray[Any]], Sequence[Path]]


def _adjust_shape_for_channels(
    shape: Sequence[int],
    channels: Sequence[int] | None,
    value: int | Literal["channels"] = "channels",
) -> tuple[int, ...]:
    """Adjust shape to account for channel subsetting.

    Parameters
    ----------
    shape : Sequence[int]
        The original data shape in SC(Z)YX format.
    channels : Sequence[int] | None
        The list of channels to select. If None, no adjustment is made.
    value : int | Literal["channels"], default="channels"
        The value to replace the channel dimension with. If "channels", the length
        of the channels list is used, by default "channels".

    Returns
    -------
    tuple[int, ...]
        The adjusted data shape in SC(Z)YX format.
    """
    if channels is not None:
        adjusted_shape = list(shape)
        adjusted_shape[1] = len(channels) if value == "channels" else value
        return tuple(adjusted_shape)
    return tuple(shape)


def _patch_size_within_data_shapes(
    data_shapes: Sequence[Sequence[int]], patch_size: Sequence[int]
) -> bool:
    """Determine whether all the data_shapes are greater or equal than the patch size.

    Parameters
    ----------
    data_shapes : Sequence[Sequence[int]]
        A sequence of data shapes. They must be in the format SC(Z)YX.
    patch_size : Sequence[int]
        A patch size that must specify the size of the patch in all the spatial
        dimensions, in the format (Z)YX.

    Returns
    -------
    bool
        If all the data shapes are greater or equal than the patch size.
    """
    smaller_than_shapes = [
        # skip sample and channel dimension in data_shape
        (np.array(patch_size) <= np.array(data_shape[2:])).all()
        for data_shape in data_shapes
    ]
    return all(smaller_than_shapes)


class CareamicsDataset(Dataset, Generic[GenericImageStack]):
    def __init__(
        self,
        data_config: NGDataConfig,
        input_extractor: PatchExtractor[GenericImageStack],
        target_extractor: PatchExtractor[GenericImageStack] | None = None,
        mask_extractor: PatchExtractor[GenericImageStack] | None = None,
    ) -> None:

        # Make sure all the image sizes are greater than the patch size for training
        data_shapes = [
            image_stack.data_shape for image_stack in input_extractor.image_stacks
        ]
        if data_config.mode != Mode.PREDICTING:
            if not isinstance(
                data_config.patching, WholePatchingConfig
            ) and not _patch_size_within_data_shapes(
                data_shapes, data_config.patching.patch_size
            ):
                raise ValueError(
                    "Not all images sizes are greater or equal than the patch size for "
                    "training and validation."
                )

        self.config = data_config

        self.input_extractor = input_extractor
        self.target_extractor = target_extractor

        self.patch_filter = (
            create_patch_filter(self.config.patch_filter)
            if self.config.patch_filter is not None
            else None
        )
        self.coord_filter = (
            create_coord_filter(self.config.coord_filter, mask=mask_extractor)
            if self.config.coord_filter is not None and mask_extractor is not None
            else None
        )
        self.patch_filter_patience = self.config.patch_filter_patience

        self.patching_strategy = create_patching_strategy(
            data_shapes=self.input_extractor.shapes,
            patching_config=self.config.patching,
        )

        resolve_normalization_config(
            norm_config=self.config.normalization,
            patching_strategy=self.patching_strategy,
            input_extractor=self.input_extractor,
            target_extractor=self.target_extractor,
            channels=self.config.channels,
        )
        self.normalization = create_normalization(self.config.normalization)

        self.transforms = self._initialize_transforms()

    def _initialize_transforms(self) -> Compose | None:
        if self.config.mode == Mode.TRAINING:
            return Compose(list(self.config.transforms))

        # TODO: add TTA
        return Compose([])

    def __len__(self):
        return self.patching_strategy.n_patches

    def _create_image_region(
        self, patch: np.ndarray, patch_spec: PatchSpecs, extractor: PatchExtractor
    ) -> ImageRegionData:
        data_idx = patch_spec["data_idx"]
        image_stack: GenericImageStack = extractor.image_stacks[data_idx]

        # adjust the number of channels in data_shape if needed
        data_shape = _adjust_shape_for_channels(
            shape=image_stack.data_shape,
            channels=self.config.channels,
        )

        # additional metadata for zarr image stacks
        if isinstance(image_stack, ZarrImageStack):
            additional_metadata = {
                "chunks": image_stack.chunks,
            }

            if image_stack.shards is not None:
                additional_metadata["shards"] = image_stack.shards
        else:
            additional_metadata = {}

        return ImageRegionData(
            data=patch,
            source=str(image_stack.source),
            dtype=str(image_stack.data_dtype),
            data_shape=data_shape,
            # TODO: should it be axes of the original image instead?
            axes=self.config.axes,
            region_spec=patch_spec,
            additional_metadata=additional_metadata,
        )

    def _extract_patches(
        self, patch_spec: PatchSpecs
    ) -> tuple[NDArray, NDArray | None]:
        """Extract input and target patches based on patch specifications."""
        input_patch = self.input_extractor.extract_channel_patch(
            data_idx=patch_spec["data_idx"],
            sample_idx=patch_spec["sample_idx"],
            channels=self.config.channels,
            coords=patch_spec["coords"],
            patch_size=patch_spec["patch_size"],
        )

        target_patch = (
            self.target_extractor.extract_channel_patch(
                data_idx=patch_spec["data_idx"],
                sample_idx=patch_spec["sample_idx"],
                # TODO does not allow selecting different channels for target
                channels=self.config.channels,
                coords=patch_spec["coords"],
                patch_size=patch_spec["patch_size"],
            )
            if self.target_extractor is not None
            else None
        )
        return input_patch, target_patch

    def _get_filtered_patch(
        self, index: int
    ) -> tuple[NDArray[Any], NDArray[Any] | None, PatchSpecs]:
        """Extract a patch that passes filtering criteria with retry logic."""
        should_filter = self.config.mode == Mode.TRAINING and (
            self.patch_filter is not None or self.coord_filter is not None
        )
        empty_patch = True
        patch_filter_patience = self.patch_filter_patience  # reset patience

        while empty_patch and patch_filter_patience > 0:
            # query patches
            patch_spec = self.patching_strategy.get_patch_spec(index)

            # filter patch based on coordinates if needed
            if should_filter and self.coord_filter is not None:
                if self.coord_filter.filter_out(patch_spec):
                    patch_filter_patience -= 1

                    # TODO should we raise an error rather than silently accept patches?
                    # if patience runs out without ever finding coordinates
                    # then we need to guard against an exist before defining
                    # input_patch and target_patch
                    if patch_filter_patience != 0:
                        continue

            input_patch, target_patch = self._extract_patches(patch_spec)

            # filter patch based on values if needed
            if should_filter and self.patch_filter is not None:
                empty_patch = self.patch_filter.filter_out(input_patch)
                patch_filter_patience -= 1  # decrease patience
            else:
                empty_patch = False

        return input_patch, target_patch, patch_spec

    def __getitem__(
        self, index: int
    ) -> Union[tuple[ImageRegionData], tuple[ImageRegionData, ImageRegionData]]:
        input_patch, target_patch, patch_spec = self._get_filtered_patch(index)

        # apply normalization
        input_patch, target_patch = self.normalization(input_patch, target_patch)

        # apply transforms
        if self.transforms is not None:
            if self.target_extractor is not None:
                input_patch, target_patch = self.transforms(input_patch, target_patch)
            else:
                # TODO: compose doesn't return None for target patch anymore
                #   so have to do this annoying if else
                (input_patch,) = self.transforms(input_patch, target_patch)
                target_patch = None

        input_data = self._create_image_region(
            patch=input_patch, patch_spec=patch_spec, extractor=self.input_extractor
        )

        if target_patch is not None and self.target_extractor is not None:
            target_data = self._create_image_region(
                patch=target_patch,
                patch_spec=patch_spec,
                extractor=self.target_extractor,
            )
            return input_data, target_data
        else:
            return (input_data,)



"""Pydantic model for the Shannon entropy patch filter."""

from typing import Literal

from .filter_config import FilterConfig


class ShannonFilterConfig(FilterConfig):
    """Pydantic model for the Shannon entropy patch filter."""

    name: Literal["shannon"] = "shannon"
    """Name of the filter."""

    threshold: float
    """Minimum Shannon entropy required to keep a patch."""



"""Callbacks module."""

__all__ = [
    "CareamicsCheckpointInfo",
    "DataStatsCallback",
    "HyperParametersCallback",
    "PredictionStoppedException",
    "PredictionWriterCallback",
    "ProgressBarCallback",
    "StopPredictionCallback",
    "create_write_strategy",
]

from .careamics_checkpoint_info_callback import CareamicsCheckpointInfo
from .data_stats_callback import DataStatsCallback
from .hyperparameters_callback import HyperParametersCallback
from .prediction_writer_callback import PredictionWriterCallback, create_write_strategy
from .progress_bar_callback import ProgressBarCallback
from .stop_prediction_callback import (
    PredictionStoppedException,
    StopPredictionCallback,
)



from collections.abc import Sequence

from .patching_strategy_protocol import PatchSpecs


class WholeSamplePatchingStrategy:
    # TODO: warn this strategy should only be used with batch size = 1
    #   for the case of multiple image stacks with different dimensions

    # TODO: docs
    def __init__(self, data_shapes: Sequence[Sequence[int]]):
        self.data_shapes = data_shapes

        self.patch_specs: list[PatchSpecs] = self._initialize_patch_specs()

    @property
    def n_patches(self) -> int:
        return len(self.patch_specs)

    def get_patch_spec(self, index: int) -> PatchSpecs:
        return self.patch_specs[index]

    # Note: this is used by the FileIterSampler
    def get_patch_indices(self, data_idx: int) -> Sequence[int]:
        """
        Get the patch indices will return patches for a specific `image_stack`.

        The `image_stack` corresponds to the given `data_idx`.

        Parameters
        ----------
        data_idx : int
            An index that corresponds to a given `image_stack`.

        Returns
        -------
        sequence of int
            A sequence of patch indices, that when used to index the `CAREamicsDataset
            will return a patch that comes from the `image_stack` corresponding to the
            given `data_idx`.
        """
        return [
            i
            for i, patch_spec in enumerate(self.patch_specs)
            if patch_spec["data_idx"] == data_idx
        ]

    def _initialize_patch_specs(self) -> list[PatchSpecs]:
        patch_specs: list[PatchSpecs] = []
        for data_idx, data_shape in enumerate(self.data_shapes):
            spatial_shape = data_shape[2:]
            for sample_idx in range(data_shape[0]):
                patch_specs.append(
                    {
                        "data_idx": data_idx,
                        "sample_idx": sample_idx,
                        "coords": tuple(0 for _ in spatial_shape),
                        "patch_size": spatial_shape,
                    }
                )
        return patch_specs



"""Noise models Pydantic configurations."""

__all__ = [
    "GaussianLikelihoodConfig",
    "GaussianMixtureNMConfig",
    "MultiChannelNMConfig",
    "NMLikelihoodConfig",
]


from .likelihood_config import GaussianLikelihoodConfig, NMLikelihoodConfig
from .noise_model_config import GaussianMixtureNMConfig, MultiChannelNMConfig



"""A writing strategy that caches tiles until a whole image is predicted."""

from collections import defaultdict
from pathlib import Path
from typing import Any

from careamics.dataset_ng.dataset import ImageRegionData
from careamics.file_io import WriteFunc
from careamics.lightning.dataset_ng.prediction import (
    stitch_single_prediction,
)

from .file_path_utils import create_write_file_path
from .write_strategy import WriteStrategy


class CachedTiles(WriteStrategy):
    """
    A write strategy that will cache tiles.

    Tiles are cached until a whole image is predicted on. Then the stitched
    prediction is saved.

    Parameters
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.

    Attributes
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.
    tile_cache : list of numpy.ndarray
        Tiles cached for stitching prediction.
    tile_info_cache : list of TileInformation
        Cached tile information for stitching prediction.
    """

    def __init__(
        self,
        write_func: WriteFunc,
        write_extension: str,
        write_func_kwargs: dict[str, Any],
    ) -> None:
        """
        A write strategy that will cache tiles.

        Tiles are cached until a whole image is predicted on. Then the stitched
        prediction is saved.

        Parameters
        ----------
        write_func : WriteFunc
            Function used to save predictions.
        write_extension : str
            Extension added to prediction file paths.
        write_func_kwargs : dict of {str: Any}
            Extra kwargs to pass to `write_func`.
        """
        super().__init__()

        self.write_func: WriteFunc = write_func
        self.write_extension: str = write_extension
        self.write_func_kwargs: dict[str, Any] = write_func_kwargs

        # where tiles will be cached until a whole image has been predicted
        self.tile_cache: dict[int, list[ImageRegionData]] = defaultdict(list)

    def write_batch(
        self,
        dirpath: Path,
        predictions: list[ImageRegionData],
    ) -> None:
        """
        Cache tiles until the last tile is predicted, then save the stitched image.

        Parameters
        ----------
        dirpath : Path
            Path to directory to save predictions to.
        predictions : list[ImageRegionData]
            Decollated predictions.
        """
        assert predictions is not None

        # cache tiles
        for tile in predictions:
            data_idx = tile.region_spec["data_idx"]
            self.tile_cache[data_idx].append(tile)

        self._write_images(dirpath)

    def _get_full_images(self) -> list[int]:
        """
        Get data indices of full images contained in the cache.

        Returns
        -------
        list of int
            Data indices of full images contained in the cache.
        """
        full_images = []
        for data_idx in self.tile_cache.keys():
            exp_n_tiles = self.tile_cache[data_idx][0].region_spec["total_tiles"]

            if len(self.tile_cache[data_idx]) == exp_n_tiles:
                full_images.append(data_idx)
            elif len(self.tile_cache[data_idx]) > exp_n_tiles:
                raise ValueError(
                    f"More tiles cached for data_idx {data_idx} than expected. "
                    f"Expected {exp_n_tiles}, found "
                    f"{len(self.tile_cache[data_idx])}."
                )

        return full_images

    def _stitch_and_write_single(
        self, dirpath: Path, tiles: list[ImageRegionData]
    ) -> None:
        """
        Stitch and write a single image from tiles.

        Parameters
        ----------
        dirpath : Path
            Path to directory to save predictions to.
        tiles : list[ImageRegionData]
            Tiles to stitch and write.
        """
        # stitch prediction
        prediction_image = stitch_single_prediction(tiles)
        source: Path = Path(tiles[0].source)

        # Handle array sources with postfix
        postfix = ""
        if source.stem == "array":
            data_idx = tiles[0].region_spec["data_idx"]
            postfix = f"_{data_idx}"

        file_path = create_write_file_path(
            dirpath=dirpath,
            file_path=source,
            write_extension=self.write_extension,
            postfix=postfix,
        )
        self.write_func(
            file_path=file_path, img=prediction_image, **self.write_func_kwargs
        )

    def _write_images(self, dirpath: Path) -> None:
        """
        Write full images from cached tiles.

        Parameters
        ----------
        dirpath : Path
            Path to directory to save predictions to.
        """
        full_images = self._get_full_images()
        for data_idx in full_images:
            tiles = self.tile_cache.pop(data_idx)
            self._stitch_and_write_single(dirpath, tiles)



from collections.abc import Sequence
from pathlib import Path
from typing import Any, Self

import numpy as np
import tifffile
from numpy.typing import DTypeLike, NDArray

from careamics.dataset.dataset_utils import reshape_array
from careamics.file_io.read import ReadFunc, read_tiff

from .image_utils.image_stack_utils import channel_slice, pad_patch, reshape_array_shape


class FileImageStack:
    """
    An ImageStack implementation for data that is coming from a file.

    The data will not be loaded until the `load` method is called. The `close` method
    can be used to remove the internal reference to the data.
    """

    def __init__(
        self,
        source: Path,
        axes: str,
        data_shape: tuple[int, ...],
        data_dtype: DTypeLike,
        read_func: ReadFunc,
        read_kwargs: dict[str, Any] | Any = None,
    ):
        self.source = source
        self.axes = axes
        self.data_shape = data_shape
        self.data_dtype = data_dtype
        self.read_func = read_func
        self.read_kwargs = read_kwargs
        self._data: NDArray | None = None

    def extract_patch(
        self,
        sample_idx: int,
        channels: Sequence[int] | None,  # `channels = None` to select all channels
        coords: Sequence[int],
        patch_size: Sequence[int],
    ) -> NDArray:
        if self._data is None:
            raise ValueError(
                "Cannot extract patch because data has not been loaded from "
                f"'{self.source}', the `load` method must be called first."
            )

        if (coord_dims := len(coords)) != (patch_dims := len(patch_size)):
            raise ValueError(
                "Patch coordinates and patch size must have the same dimensions but "
                f"found {coord_dims} and {patch_dims}."
            )

        # check that channels are within bounds
        if channels is not None:
            max_channel = self.data_shape[1] - 1  # channel is second dimension
            for ch in channels:
                if ch > max_channel:
                    raise ValueError(
                        f"Channel index {ch} is out of bounds for data with "
                        f"{self.data_shape[1]} channels. Check the provided `channels` "
                        f"parameter in the configuration for erroneous channel "
                        f"indices."
                    )

        patch_data = self._data[
            (
                sample_idx,  # type: ignore
                # use channel slice so that channel dimension is kept
                channel_slice(channels),  # type: ignore
                *[
                    slice(
                        np.clip(c, 0, self.data_shape[2 + i]),
                        np.clip(c + ps, 0, self.data_shape[2 + i]),
                    )
                    for i, (c, ps) in enumerate(zip(coords, patch_size, strict=False))
                ],  # type: ignore
            )  # type: ignore
        ]
        patch = pad_patch(coords, patch_size, self.data_shape, patch_data)

        return patch

    def load(self):
        """Load the data stored in a file."""
        data = self.read_func(self.source)
        self._data = reshape_array(data, self.axes)

    # TODO: maybe this should be called something else
    def close(self):
        """Remove the internal reference to the data to clear up memory."""
        # will get cleaned up by the garbage collector since there is no longer a ref
        self._data = None

    @property
    def is_loaded(self):
        return self._data is not None

    @classmethod
    def from_tiff(
        cls,
        path: Path,
        axes: str,
    ) -> Self:
        """
        Construct the `ImageStack` from a TIFF file.

        Parameters
        ----------
        path : Path
            Path to the TIFF file.
        axes : str
            The original axes of the data, must be a subset of STCZYX.

        Returns
        -------
        Self
            The `ImageStack` with the underlying data being from a TIFF file.
        """
        # TODO: think this is correct but need more examples to test
        file = tifffile.TiffFile(path)
        data_shape = reshape_array_shape(axes, file.series[0].shape)
        dtype = file.series[0].dtype
        return cls(
            source=path,
            axes=axes,
            data_shape=data_shape,
            data_dtype=dtype,
            read_func=read_tiff,
        )



from pathlib import Path

import numpy as np
import pytest

from careamics.config import (
    CAREAlgorithm,
    Configuration,
    MicroSplitAlgorithm,
    N2NAlgorithm,
    N2VAlgorithm,
    PN2VAlgorithm,
    algorithm_factory,
    create_care_configuration,
    create_microsplit_configuration,
    create_n2n_configuration,
    create_n2v_configuration,
    create_pn2v_configuration,
)
from careamics.config.configuration_factories import (
    _create_data_configuration,
    _create_supervised_config_dict,
    _create_unet_configuration,
    _list_spatial_augmentations,
)
from careamics.config.support import (
    SupportedAlgorithm,
    SupportedPixelManipulation,
    SupportedStructAxis,
    SupportedTransform,
)
from careamics.config.transformations import (
    N2VManipulateConfig,
    XYFlipConfig,
    XYRandomRotate90Config,
)


def test_careamics_config_n2v(minimum_n2v_configuration):
    """Test that the N2V configuration is created correctly."""
    config = Configuration(**minimum_n2v_configuration)
    assert config.algorithm_config.algorithm == SupportedAlgorithm.N2V.value


@pytest.mark.parametrize(
    "algorithm", [SupportedAlgorithm.N2N.value, SupportedAlgorithm.CARE.value]
)
def test_careamics_config_supervised(minimum_supervised_configuration, algorithm):
    """Test that the supervised configuration is created correctly."""
    min_config = minimum_supervised_configuration
    min_config["algorithm_config"]["algorithm"] = algorithm

    config = Configuration(**min_config)

    assert config.algorithm_config.algorithm == algorithm


def test_algorithm_factory_n2v(minimum_algorithm_n2v):
    """Test that the N2V configuration is created correctly."""
    algorithm = algorithm_factory(minimum_algorithm_n2v)
    assert isinstance(algorithm, N2VAlgorithm)


@pytest.mark.parametrize("algorithm", ["n2n", "care"])
def test_algorithm_factory_supervised(minimum_algorithm_supervised, algorithm):
    """Test that the supervised configuration is created correctly."""
    min_config = minimum_algorithm_supervised
    min_config["algorithm"] = algorithm

    algorithm_config = algorithm_factory(min_config)

    exp_class = N2NAlgorithm if algorithm == "n2n" else CAREAlgorithm
    assert isinstance(algorithm_config, exp_class)


def test_list_aug_default():
    """Test that the default augmentations are present."""
    list_aug = _list_spatial_augmentations(augmentations=None)

    assert len(list_aug) == 2
    assert list_aug[0].name == SupportedTransform.XY_FLIP.value
    assert list_aug[1].name == SupportedTransform.XY_RANDOM_ROTATE90.value


def test_list_aug_no_aug():
    """Test that disabling augmentation results in empty transform list."""
    list_aug = _list_spatial_augmentations(augmentations=[])
    assert len(list_aug) == 0


def test_list_aug_error_duplicate_transforms():
    """Test that an error is raised when there are duplicate transforms."""
    with pytest.raises(ValueError):
        _list_spatial_augmentations(
            augmentations=[XYFlipConfig(), XYRandomRotate90Config(), XYFlipConfig()],
        )


def test_list_aug_error_wrong_transform():
    """Test that an error is raised when the wrong transform is passed."""
    with pytest.raises(ValueError):
        _list_spatial_augmentations(
            augmentations=[XYFlipConfig(), N2VManipulateConfig()],
        )


def test_create_data_configuration_train_dataloader_params(minimum_data):
    """Test that shuffle is added silently to the train_dataloader_params."""
    config_dict = minimum_data
    config_dict["train_dataloader_params"] = {"num_workers": 4}

    config = _create_data_configuration(batch_size=1, augmentations=[], **config_dict)
    assert "shuffle" in config.train_dataloader_params
    assert config.train_dataloader_params["shuffle"]


def test_supervised_configuration_passing_transforms():
    """Test that transforms can be passed to the configuration."""
    config_dict = _create_supervised_config_dict(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        augmentations=[XYFlipConfig()],
    )
    config = Configuration(**config_dict)

    assert len(config.data_config.transforms) == 1
    assert config.data_config.transforms[0].name == SupportedTransform.XY_FLIP.value


def test_model_creation():
    """Test that the correct parameters are passed to the model."""
    model_kwargs = {
        "depth": 4,
        "conv_dims": 2,
        "n2v2": False,
        "in_channels": 2,
        "num_classes": 5,
        "independent_channels": False,
    }

    # choose different parameters
    axes = "XYZ"
    conv_dims = 3
    in_channels = 3
    num_classes = 4
    independent_channels = True
    use_n2v2 = True

    model = _create_unet_configuration(
        axes=axes,
        n_channels_in=in_channels,
        n_channels_out=num_classes,
        independent_channels=independent_channels,
        use_n2v2=use_n2v2,
        model_params=model_kwargs,
    )

    assert model.depth == model_kwargs["depth"]
    assert model.conv_dims == conv_dims
    assert model.n2v2 == use_n2v2
    assert model.in_channels == in_channels
    assert model.num_classes == num_classes
    assert model.independent_channels == independent_channels


def test_create_configuration():
    """Test that the methods correctly passes all parameters."""
    algorithm = "care"
    experiment_name = "test"
    data_type = "tiff"
    axes = "CYX"
    patch_size = [128, 128]
    batch_size = 8
    transform_list = [XYFlipConfig(), XYRandomRotate90Config()]
    independent_channels = False
    loss = "mse"
    n_channels_in = 2
    n_channels_out = 3
    logger = "tensorboard"
    model_params = {
        "depth": 5,
    }
    optimizer = "SGD"
    optimizer_params = {
        "lr": 0.07,
    }
    lr_scheduler = "StepLR"
    lr_scheduler_params = {
        "step_size": 19,
    }
    train_dataloader_params = {
        "num_workers": 4,
        "shuffle": True,
    }
    val_dataloader_params = {
        "num_workers": 4,
    }
    checkpoint_params = {
        "every_n_epochs": 9,
    }

    # instantiate config
    config = Configuration(
        **_create_supervised_config_dict(
            algorithm=algorithm,
            experiment_name=experiment_name,
            data_type=data_type,
            axes=axes,
            patch_size=patch_size,
            batch_size=batch_size,
            augmentations=transform_list,
            independent_channels=independent_channels,
            loss=loss,
            n_channels_in=n_channels_in,
            n_channels_out=n_channels_out,
            logger=logger,
            model_params=model_params,
            optimizer=optimizer,
            optimizer_params=optimizer_params,
            lr_scheduler=lr_scheduler,
            lr_scheduler_params=lr_scheduler_params,
            train_dataloader_params=train_dataloader_params,
            val_dataloader_params=val_dataloader_params,
            checkpoint_params=checkpoint_params,
        )
    )

    assert config.algorithm_config.algorithm == algorithm
    assert config.experiment_name == experiment_name
    assert config.data_config.data_type == data_type
    assert config.data_config.axes == axes
    assert config.data_config.patch_size == patch_size
    assert config.data_config.batch_size == batch_size
    assert config.data_config.transforms == transform_list
    assert config.algorithm_config.model.independent_channels == independent_channels
    assert config.algorithm_config.loss == loss
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_out
    assert config.training_config.logger == logger
    assert config.algorithm_config.model.depth == model_params["depth"]
    assert config.algorithm_config.optimizer.name == optimizer
    assert config.algorithm_config.optimizer.parameters == optimizer_params
    assert config.algorithm_config.lr_scheduler.name == lr_scheduler
    assert config.algorithm_config.lr_scheduler.parameters == lr_scheduler_params
    assert config.data_config.train_dataloader_params == train_dataloader_params
    assert config.data_config.val_dataloader_params == val_dataloader_params
    assert (
        config.training_config.checkpoint_callback.every_n_epochs
        == checkpoint_params["every_n_epochs"]
    )


def test_supervised_configuration_error_with_channel_axes():
    """Test that an error is raised if channels are in axes, but the input channel
    number is not specified."""
    with pytest.raises(ValueError):
        _create_supervised_config_dict(
            algorithm="n2n",
            experiment_name="test",
            data_type="tiff",
            axes="CYX",
            patch_size=[64, 64],
            batch_size=8,
        )


def test_supervised_configuration_singleton_channel():
    """Test that no error is raised if channels are in axes, and the input channel is
    1."""
    _create_supervised_config_dict(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        n_channels_in=1,
    )


def test_supervised_configuration_no_channel():
    """Test that no error is raised without channel and number of inputs."""
    _create_supervised_config_dict(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
    )


def test_supervised_configuration_error_without_channel_axes():
    """Test that an error is raised if channels are not in axes, but the input channel
    number is specified and greater than 1."""
    with pytest.raises(ValueError):
        _create_supervised_config_dict(
            algorithm="n2n",
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            n_channels_in=2,
        )


def test_supervised_configuration_channels():
    """Test that no error is raised if channels are in axes and the input channel
    are specified."""
    _create_supervised_config_dict(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        n_channels_in=4,
    )


def test_n2n_configuration():
    """Test that N2N configuration can be created."""
    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        train_dataloader_params={"num_workers": 2},
    )
    assert config.algorithm_config.algorithm == "n2n"


def test_n2n_configuration_n_channels():
    """Test the behaviour of the number of channels in and out."""
    n_channels_in = 4
    n_channels_out = 5

    # n_channels_out not specified
    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        n_channels_in=n_channels_in,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_in

    # specify n_channels_out
    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_out


def test_n2n_configuration_limit_train_batches():
    """Num_steps parameter is passed to trainer config as limit_train_batches."""
    num_steps = 15

    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_steps=num_steps,
    )
    assert (
        config.training_config.lightning_trainer_config["limit_train_batches"]
        == num_steps
    )

    # Test with float value
    num_steps_float = 0.3
    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_steps=num_steps_float,
    )
    assert (
        config.training_config.lightning_trainer_config["limit_train_batches"]
        == num_steps_float
    )

    # Test without num_steps (should not be in config)
    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
    )
    assert "limit_train_batches" not in config.training_config.lightning_trainer_config


def test_care_configuration():
    """Test that CARE configuration can be created."""
    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        train_dataloader_params={"num_workers": 2},
    )
    assert config.algorithm_config.algorithm == "care"


def test_care_configuration_n_channels():
    """Test the behaviour of the number of channels in and out."""
    n_channels_in = 4
    n_channels_out = 5

    # n_channels_out not specified
    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        n_channels_in=n_channels_in,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_in

    # specify n_channels_out
    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_out


def test_care_configuration_limit_train_batches():
    """Num_steps parameter is passed to trainer config as limit_train_batches."""
    num_steps = 10

    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_steps=num_steps,
    )
    assert (
        config.training_config.lightning_trainer_config["limit_train_batches"]
        == num_steps
    )

    # Test with float value
    num_steps_float = 0.5
    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_steps=num_steps_float,
    )
    assert (
        config.training_config.lightning_trainer_config["limit_train_batches"]
        == num_steps_float
    )

    # Test without num_steps (should not be in config)
    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
    )
    assert "limit_train_batches" not in config.training_config.lightning_trainer_config


def test_care_configuration_trainer_params():
    """Test that trainer_params are correctly passed to CARE trainer config."""
    num_epochs = 40
    trainer_params = {
        "accelerator": "gpu",
        "devices": 1,
        "precision": "16-mixed",
        "gradient_clip_val": 0.8,
        "check_val_every_n_epoch": 3,
        "deterministic": True,
    }

    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=num_epochs,
        trainer_params=trainer_params,
    )

    assert config.training_config.lightning_trainer_config["max_epochs"] == num_epochs
    for key, value in trainer_params.items():
        assert config.training_config.lightning_trainer_config[key] == value


def test_care_configuration_trainer_params_override():
    """Test that explicit parameters override trainer_params in CARE configuration."""
    num_epochs = 35
    num_steps = 750

    trainer_params = {
        "max_epochs": 200,
        "limit_train_batches": 3000,
        "precision": "32-true",
    }

    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=num_epochs,
        num_steps=num_steps,
        trainer_params=trainer_params,
    )

    # Explicit parameters should override
    assert config.training_config.lightning_trainer_config["max_epochs"] == num_epochs
    assert (
        config.training_config.lightning_trainer_config["limit_train_batches"]
        == num_steps
    )
    # Non-conflicting should be preserved
    assert config.training_config.lightning_trainer_config["precision"] == "32-true"


def test_n2v_configuration():
    """Test that N2V configuration can be created."""
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        train_dataloader_params={"num_workers": 2},
    )
    assert isinstance(config.algorithm_config, N2VAlgorithm)


def test_n2v_configuration_no_aug():
    """Test the default n2v transforms."""
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        augmentations=[],
    )
    assert config.data_config.transforms == []


def test_n2v_configuration_n2v2_structn2v():
    """Test that N2V manipulate is ignored when explicitely passed."""
    use_n2v2 = True
    roi_size = 15
    masked_pixel_percentage = 0.5
    struct_mask_axis = SupportedStructAxis.HORIZONTAL.value
    struct_n2v_span = 15

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        use_n2v2=use_n2v2,  # median strategy
        roi_size=roi_size,
        masked_pixel_percentage=masked_pixel_percentage,
        struct_n2v_axis=struct_mask_axis,
        struct_n2v_span=struct_n2v_span,
    )
    assert (
        config.algorithm_config.n2v_config.strategy
        == SupportedPixelManipulation.MEDIAN.value
    )
    assert config.algorithm_config.n2v_config.roi_size == roi_size
    assert (
        config.algorithm_config.n2v_config.masked_pixel_percentage
        == masked_pixel_percentage
    )
    assert config.algorithm_config.n2v_config.struct_mask_axis == struct_mask_axis
    assert config.algorithm_config.n2v_config.struct_mask_span == struct_n2v_span


def test_n2v_configuration_limit_train_batches():
    """Num_steps parameter is passed to trainer config as limit_train_batches."""
    num_steps = 20

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_steps=num_steps,
    )
    assert (
        config.training_config.lightning_trainer_config["limit_train_batches"]
        == num_steps
    )

    # Test with float value
    num_steps_float = 0.7
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_steps=num_steps_float,
    )
    assert (
        config.training_config.lightning_trainer_config["limit_train_batches"]
        == num_steps_float
    )

    # Test without num_steps (should not be in config)
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
    )
    assert "limit_train_batches" not in config.training_config.lightning_trainer_config


def test_n2v_configuration_num_epochs():
    """Test that num_epochs parameter is correctly passed to trainer config."""
    num_epochs = 50

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=num_epochs,
    )
    assert config.training_config.lightning_trainer_config["max_epochs"] == num_epochs

    # Test with num_epochs=None (should not be in config)
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=None,
    )
    assert "max_epochs" not in config.training_config.lightning_trainer_config


def test_n2v_configuration_num_steps():
    """Test that num_steps parameter is correctly passed to trainer config."""
    num_steps = 1000

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_steps=num_steps,
    )
    assert (
        config.training_config.lightning_trainer_config["limit_train_batches"]
        == num_steps
    )

    # Test without num_steps (should not be in config)
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
    )
    assert "limit_train_batches" not in config.training_config.lightning_trainer_config


def test_n2v_configuration_num_epochs_and_num_steps():
    """Test that both num_epochs and num_steps can be set simultaneously."""
    num_epochs = 25
    num_steps = 500

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=num_epochs,
        num_steps=num_steps,
    )
    assert config.training_config.lightning_trainer_config["max_epochs"] == num_epochs
    assert (
        config.training_config.lightning_trainer_config["limit_train_batches"]
        == num_steps
    )


def test_n2v_configuration_trainer_params():
    """Test that various trainer_params are correctly passed to trainer config."""
    trainer_params = {
        "accelerator": "gpu",
        "devices": 2,
        "precision": 16,
        "gradient_clip_val": 0.5,
        "accumulate_grad_batches": 4,
        "check_val_every_n_epoch": 5,
        "log_every_n_steps": 10,
        "enable_checkpointing": True,
        "enable_progress_bar": False,
        "enable_model_summary": True,
    }

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        trainer_params=trainer_params,
    )

    for key, value in trainer_params.items():
        assert config.training_config.lightning_trainer_config[key] == value


def test_n2v_configuration_trainer_params_with_timing():
    """Test trainer_params with timing-related parameters."""
    trainer_params = {
        "min_epochs": 5,
        "min_steps": 100,
        "max_time": "00:01:00:00",  # 1 hour
        "val_check_interval": 0.25,
    }

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        trainer_params=trainer_params,
    )

    for key, value in trainer_params.items():
        assert config.training_config.lightning_trainer_config[key] == value


def test_n2v_configuration_trainer_params_override():
    """Test that explicit parameters override trainer_params."""
    num_epochs = 30
    num_steps = 800

    # trainer_params has conflicting values
    trainer_params = {
        "max_epochs": 100,
        "limit_train_batches": 0.5,
        "accelerator": "cpu",
    }

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=num_epochs,
        num_steps=num_steps,
        trainer_params=trainer_params,
    )

    # Explicit parameters should override trainer_params
    assert config.training_config.lightning_trainer_config["max_epochs"] == num_epochs
    assert (
        config.training_config.lightning_trainer_config["limit_train_batches"]
        == num_steps
    )

    # Non-conflicting trainer_params should be preserved
    assert config.training_config.lightning_trainer_config["accelerator"] == "cpu"


def test_n2v_configuration_trainer_params_profiler():
    """Test trainer_params with profiler settings."""
    trainer_params = {
        "profiler": "simple",
        "detect_anomaly": True,
        "benchmark": False,
        "deterministic": True,
    }

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        trainer_params=trainer_params,
    )

    for key, value in trainer_params.items():
        assert config.training_config.lightning_trainer_config[key] == value


def test_n2v_configuration_trainer_params_distributed():
    """Test trainer_params with distributed training settings."""
    trainer_params = {
        "strategy": "ddp",
        "num_nodes": 2,
        "sync_batchnorm": True,
        "use_distributed_sampler": True,
    }

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        trainer_params=trainer_params,
    )

    for key, value in trainer_params.items():
        assert config.training_config.lightning_trainer_config[key] == value


def test_n2v_configuration_all_trainer_combinations():
    """Test comprehensive combination of all trainer parameters."""
    num_epochs = 15
    num_steps = 300

    trainer_params = {
        "accelerator": "auto",
        "devices": "auto",
        "precision": "32-true",
        "gradient_clip_val": 1.0,
        "gradient_clip_algorithm": "norm",
        "accumulate_grad_batches": 2,
        "check_val_every_n_epoch": 2,
        "val_check_interval": 1.0,
        "log_every_n_steps": 50,
        "enable_checkpointing": True,
        "enable_progress_bar": True,
        "enable_model_summary": True,
        "deterministic": False,
        "benchmark": True,
        "inference_mode": True,
        "use_distributed_sampler": True,
        "detect_anomaly": False,
        "barebones": False,
        "reload_dataloaders_every_n_epochs": 0,
    }

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=num_epochs,
        num_steps=num_steps,
        trainer_params=trainer_params,
    )

    # Check explicit parameters
    assert config.training_config.lightning_trainer_config["max_epochs"] == num_epochs
    assert (
        config.training_config.lightning_trainer_config["limit_train_batches"]
        == num_steps
    )

    # Check trainer_params
    for key, value in trainer_params.items():
        assert config.training_config.lightning_trainer_config[key] == value


def test_n2v_configuration_empty_trainer_params():
    """Test that empty trainer_params dict works correctly."""
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=None,
        trainer_params={},
    )

    # Should have empty dict (no trainer params)
    assert config.training_config.lightning_trainer_config == {}


def test_n2v_configuration_trainer_params_none():
    """Test that trainer_params=None works correctly (default behavior)."""
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=None,
        trainer_params=None,
    )

    # Should have empty dict when trainer_params is None
    assert config.training_config.lightning_trainer_config == {}


def test_pn2v_configuration(tmp_path: Path, create_dummy_noise_model):
    """Test that PN2V configuration can be created."""
    np.savez(tmp_path / "dummy_noise_model.npz", **create_dummy_noise_model)

    config = create_pn2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        nm_path=tmp_path / "dummy_noise_model.npz",
        train_dataloader_params={"num_workers": 2},
    )
    assert isinstance(config.algorithm_config, PN2VAlgorithm)


def test_checkpoint_model_save_top_k_default():
    """CheckpointConfig's default save_top_k=3 doesn't conflict with trainer params."""
    from careamics.config.lightning.callbacks.callback_config import CheckpointConfig

    # Test default save_top_k value
    checkpoint_model = CheckpointConfig()
    assert checkpoint_model.save_top_k == 3

    # Test that default works with standard trainer configuration
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=10,
        checkpoint_params={"save_top_k": 3},  # explicit default
    )

    assert config.training_config.checkpoint_callback.save_top_k == 3


def test_checkpoint_model_comprehensive_checkpoint_params():
    """Test comprehensive checkpoint parameters with save_top_k=3 default."""
    checkpoint_params = {
        "save_top_k": 3,  # explicit default
        "monitor": "val_loss",
        "mode": "min",
        "save_weights_only": False,
        "save_last": True,
        "every_n_epochs": 1,
        "verbose": True,
    }

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=10,
        checkpoint_params=checkpoint_params,
    )

    # Verify all checkpoint parameters are set correctly
    checkpoint_callback = config.training_config.checkpoint_callback
    assert checkpoint_callback.save_top_k == 3
    assert checkpoint_callback.monitor == "val_loss"
    assert checkpoint_callback.mode == "min"
    assert checkpoint_callback.save_weights_only is False
    assert checkpoint_callback.save_last is True
    assert checkpoint_callback.every_n_epochs == 1
    assert checkpoint_callback.verbose is True


def test_checkpoint_model_save_top_k_with_resource_constraints():
    """Test checkpoint save_top_k behavior with resource constraint scenarios."""
    # Test with limited training steps - checkpointing should still work
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_steps=50,
        checkpoint_params={"save_top_k": 3, "every_n_train_steps": 10},
    )

    assert config.training_config.lightning_trainer_config["limit_train_batches"] == 50
    assert config.training_config.checkpoint_callback.save_top_k == 3
    assert config.training_config.checkpoint_callback.every_n_train_steps == 10


def test_checkpoint_model_save_top_k_edge_cases():
    """Test edge cases for save_top_k parameter."""
    import pytest

    from careamics.config.lightning.callbacks.callback_config import CheckpointConfig

    # Test that save_top_k accepts valid range (-1 to 100)
    checkpoint_model = CheckpointConfig(save_top_k=-1)  # Save all
    assert checkpoint_model.save_top_k == -1

    checkpoint_model = CheckpointConfig(save_top_k=0)  # Save none
    assert checkpoint_model.save_top_k == 0

    checkpoint_model = CheckpointConfig(save_top_k=100)  # Maximum allowed
    assert checkpoint_model.save_top_k == 100

    # Test that invalid values are rejected (assuming validation exists)
    # This depends on the Pydantic field constraints
    with pytest.raises(ValueError):
        CheckpointConfig(save_top_k=101)  # Above maximum

    with pytest.raises(ValueError):
        CheckpointConfig(save_top_k=-2)  # Below minimum


def test_checkpoint_model_save_top_k_with_development_modes():
    """Test checkpoint save_top_k with various development and debugging modes."""
    # Test with overfit_batches (overfitting mode)
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        trainer_params={"overfit_batches": 1},
        checkpoint_params={"save_top_k": 3},
    )
    assert config.training_config.lightning_trainer_config["overfit_batches"] == 1
    assert config.training_config.checkpoint_callback.save_top_k == 3

    # Test with limit_val_batches
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        trainer_params={"limit_val_batches": 0.1},
        checkpoint_params={"save_top_k": 3},
    )
    assert config.training_config.lightning_trainer_config["limit_val_batches"] == 0.1
    assert config.training_config.checkpoint_callback.save_top_k == 3

    # Test with num_sanity_val_steps=0 (disable sanity checks)
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        trainer_params={"num_sanity_val_steps": 0},
        checkpoint_params={"save_top_k": 3},
    )
    assert config.training_config.lightning_trainer_config["num_sanity_val_steps"] == 0
    assert config.training_config.checkpoint_callback.save_top_k == 3


def test_checkpoint_model_save_top_k_cross_algorithm_consistency():
    """Test that save_top_k behavior is consistent across all three algorithms."""
    checkpoint_params = {"save_top_k": 5, "monitor": "val_loss"}

    # Test N2V
    n2v_config = create_n2v_configuration(
        experiment_name="test_n2v",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=10,
        checkpoint_params=checkpoint_params,
    )

    # Test CARE
    care_config = create_care_configuration(
        experiment_name="test_care",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=10,
        checkpoint_params=checkpoint_params,
    )

    # Test N2N
    n2n_config = create_n2n_configuration(
        experiment_name="test_n2n",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=10,
        checkpoint_params=checkpoint_params,
    )

    # All should have the same checkpoint configuration
    assert n2v_config.training_config.checkpoint_callback.save_top_k == 5
    assert care_config.training_config.checkpoint_callback.save_top_k == 5
    assert n2n_config.training_config.checkpoint_callback.save_top_k == 5

    assert n2v_config.training_config.checkpoint_callback.monitor == "val_loss"
    assert care_config.training_config.checkpoint_callback.monitor == "val_loss"
    assert n2n_config.training_config.checkpoint_callback.monitor == "val_loss"


def test_microsplit_configuration(tmp_path: Path, create_dummy_noise_model):
    """Test that MicroSplit configuration can be created."""
    np.savez(tmp_path / "dummy_noise_model.npz", **create_dummy_noise_model)

    config = create_microsplit_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        predict_logvar="pixelwise",
        nm_paths=[tmp_path / "dummy_noise_model.npz"],
        data_stats=[0, 0],
        train_dataloader_params={"num_workers": 0},
    )
    assert config.algorithm_config.algorithm == "microsplit"
    assert isinstance(config.algorithm_config, MicroSplitAlgorithm)
    assert config.algorithm_config.model.architecture == "LVAE"
    assert config.algorithm_config.noise_model_likelihood is not None



import itertools
from collections.abc import Sequence

import numpy as np
import pytest

from careamics.dataset_ng.patching_strategies import StratifiedPatchingStrategy


@pytest.mark.parametrize(
    "data_shapes,patch_size",
    [
        [[(2, 1, 64, 64), (1, 1, 43, 71), (3, 1, 16, 9)], (8, 8)],
        [[(2, 1, 64, 64), (1, 1, 43, 71), (3, 1, 14, 9)], (8, 5)],
        [[(2, 1, 64, 64, 64), (1, 1, 43, 71, 46), (3, 1, 16, 9, 12)], (8, 8, 8)],
        [[(2, 1, 64, 64, 64), (1, 1, 43, 71, 46), (3, 1, 14, 9, 12)], (8, 5, 7)],
    ],
)
def test_excluded_patches_never_selected(
    data_shapes: Sequence[Sequence[int]], patch_size: Sequence[int]
):
    """Ensure excluded patches are never sampled from."""
    rng = np.random.default_rng(42)
    patching_strategy = StratifiedPatchingStrategy(data_shapes, patch_size, 42)

    excluded_masks = [np.zeros(shape, dtype=bool) for shape in data_shapes]
    tracking_arrays = [np.zeros(shape, dtype=int) for shape in data_shapes]

    # exclude patches, choose randomly
    for data_idx, data_shape in enumerate(data_shapes):
        for sample_idx in range(data_shape[0]):
            grid_shape = patching_strategy.image_patching[data_idx][
                sample_idx
            ].grid_shape
            # exclude 1/8 th of patches
            n = int(np.ceil(np.prod(grid_shape) / 8))
            indices = rng.choice(np.prod(grid_shape), n)
            grid_coords = list(itertools.product(*[range(s) for s in grid_shape]))
            selected = [grid_coords[idx] for idx in indices]

            patching_strategy.exclude_patches(data_idx, sample_idx, selected)
            for grid_coord in selected:

                exluded_slice = [
                    slice(c * ps, (c + 1) * ps)
                    for c, ps in zip(grid_coord, patch_size, strict=True)
                ]
                excluded_masks[data_idx][sample_idx, ..., *exluded_slice] = True

    epochs = 5
    for _ in range(epochs):
        for i in range(patching_strategy.n_patches):
            patch_spec = patching_strategy.get_patch_spec(i)
            data_idx = patch_spec["data_idx"]
            sample_idx = patch_spec["sample_idx"]
            coord = patch_spec["coords"]
            patch_slice = [
                slice(c, c + ps) for c, ps in zip(coord, patch_size, strict=True)
            ]
            tracking_arrays[data_idx][sample_idx, ..., *patch_slice] += 1

    for excluded_mask, tracking_array in zip(
        excluded_masks, tracking_arrays, strict=True
    ):
        assert (tracking_array[excluded_mask] == 0).all()
        # for good measure make sure the rest of the array isn't zeros
        assert (tracking_array[~excluded_mask] != 0).any()



"""Test PN2V loss function."""

from pathlib import Path

import numpy as np
import torch

from careamics.config.noise_model import GaussianMixtureNMConfig
from careamics.losses.fcn.losses import pn2v_loss
from careamics.models.lvae.noise_models import GaussianMixtureNoiseModel


def test_pn2v_loss_basic(tmp_path: Path, create_dummy_noise_model):
    """Test that PN2V loss function works with basic inputs."""
    np.savez(tmp_path / "dummy_noise_model.npz", **create_dummy_noise_model)

    batch_size, height, width = 2, 32, 32

    samples = torch.randn(batch_size, 1, height, width)
    labels = torch.randn(batch_size, 1, height, width)
    masks = torch.randint(0, 2, (batch_size, 1, height, width)).float()

    # Create a noise model config
    nm_config = GaussianMixtureNMConfig(
        model_type="GaussianMixtureNoiseModel", path=tmp_path / "dummy_noise_model.npz"
    )

    # TODO Create noise model (we'll need to mock this properly)
    # For now, create a simple mock
    noise_model = GaussianMixtureNoiseModel(nm_config)

    # Test that loss function can be called
    try:
        loss_value = pn2v_loss(samples, labels, masks, noise_model)
        assert isinstance(loss_value, torch.Tensor)
        assert loss_value.dim() == 0  # Should be a scalar
        print(f"PN2V loss computed successfully: {loss_value.item()}")
    except Exception as e:
        print(f"Error calling pn2v_loss: {e}")
        raise



from pathlib import Path

import numpy as np
import pytest
import tifffile
import zarr
from numpy.typing import NDArray

import careamics.lightning.dataset_ng.callbacks.prediction_writer as pd_writer
from careamics.config.data import NGDataConfig
from careamics.dataset.dataset_utils import reshape_array
from careamics.dataset_ng.dataset import CareamicsDataset, ImageRegionData
from careamics.dataset_ng.image_stack_loader import load_arrays, load_tiffs, load_zarrs
from careamics.dataset_ng.patch_extractor import PatchExtractor
from careamics.dataset_ng.patching_strategies import (
    PatchSpecs,
    TileSpecs,
    TilingStrategy,
)

# to comply with ruff line length
WriteTilesZarr = pd_writer.write_tiles_zarr_strategy.WriteTilesZarr
_auto_chunks = pd_writer.write_tiles_zarr_strategy._auto_chunks
_update_data_shape = pd_writer.write_tiles_zarr_strategy._update_data_shape

# TODO test chunking and sharding errors and handling (e.g. when missing)


def create_image_region(
    axes, patch: np.ndarray, patch_spec: PatchSpecs, extractor: PatchExtractor
) -> ImageRegionData:
    data_idx = patch_spec["data_idx"]
    source = extractor.image_stacks[data_idx].source
    shards = (
        extractor.image_stacks[data_idx].shards
        if hasattr(extractor.image_stacks[data_idx], "shards")
        else None
    )
    chunks = (
        extractor.image_stacks[data_idx].chunks
        if hasattr(extractor.image_stacks[data_idx], "chunks")
        else None
    )
    return ImageRegionData(
        data=patch,
        source=str(source),
        dtype=str(extractor.image_stacks[data_idx].data_dtype),
        data_shape=extractor.image_stacks[data_idx].data_shape,
        axes=axes,
        region_spec=patch_spec,
        additional_metadata={
            "shards": shards,
            "chunks": chunks,
        },
    )


def gen_image_regions(
    axes: str, my_patch_extractor: PatchExtractor, my_strategy: TilingStrategy
):
    for i in range(my_strategy.n_patches):
        patch_spec: TileSpecs = my_strategy.get_patch_spec(i)
        patch = my_patch_extractor.extract_patch(
            data_idx=patch_spec["data_idx"],
            sample_idx=patch_spec["sample_idx"],
            coords=patch_spec["coords"],
            patch_size=patch_spec["patch_size"],
        )
        region = create_image_region(axes, patch, patch_spec, my_patch_extractor)

        yield region


@pytest.fixture
def data_config(axes, shape, channels) -> NGDataConfig:
    # create tiling strategy
    if "Z" in axes:
        tile_size = (8, 16, 16)
        overlaps = (2, 4, 4)
    else:
        tile_size = (16, 16)
        overlaps = (4, 4)

    if "C" in axes:
        n_channels = shape[axes.index("C")]

        if channels is not None:
            n_channels = len(channels)
    else:
        n_channels = 1

    return NGDataConfig(
        mode="predicting",
        data_type="zarr",
        patching={
            "name": "tiled",
            "patch_size": tile_size,
            "overlaps": overlaps,
        },
        axes=axes,
        channels=channels,
        normalization={
            "name": "mean_std",
            "input_means": [0.0 for _ in range(n_channels)],
            "input_stds": [1.0 for _ in range(n_channels)],
        },
    )


# TODO this is very similar to the fixture in test_ng_stitch_prediction.py, refactor
@pytest.fixture
def tiles(
    tmp_path, data_config: NGDataConfig, n_data, shape, shards, chunks
) -> tuple[NDArray, list[ImageRegionData]]:
    """Create tiles.

    Note that the tiles will be slightly different because the dataset changes the dtype
    of the data and performs normalization.

    We must use np.testing.assert_allclose(stitched_array, array, rtol=1e-5, atol=0),
    with relative tolerance as the errors scale with the values and we use
    np.arange to create the data.

    Returns
    -------
    np.ndarray
        Original array of shape DSC(Z)YX, where D is data.
    list of ImageRegionData
        Extracted tiles.
    """
    # create data
    array = np.arange(n_data * np.prod(shape)).reshape((n_data, *shape))

    sources = []
    root = tmp_path / "input_data"
    root.mkdir(parents=True, exist_ok=True)

    even = root / "even.zarr"
    odd = root / "odd.zarr"

    for i in range(n_data):
        if i % 2 == 0:
            pth = even
        else:
            pth = odd

        if not pth.exists():
            g = zarr.create_group(pth)
        else:
            g = zarr.open(pth, mode="a")

        # write array
        arr = g.create_array(
            name=f"array_{i}",
            data=array[i],
            shards=shards,
            chunks=chunks,
        )
        sources.append(arr.store_path)

    if "S" in data_config.axes:
        if "C" in data_config.axes:
            shape_with_sc = shape
        else:
            shape_with_sc = (shape[0], 1, *shape[1:])
    else:
        if "C" in data_config.axes:
            shape_with_sc = (1, *shape)
        else:
            shape_with_sc = (1, 1, *shape)

    tiling_strategy = TilingStrategy(
        data_shapes=[shape_with_sc] * n_data,
        patch_size=data_config.patching.patch_size,
        overlaps=data_config.patching.overlaps,
    )
    n_tiles = tiling_strategy.n_patches

    # create patch extractor
    image_stacks = load_zarrs(source=sources, axes=data_config.axes)
    patch_extractor = PatchExtractor(image_stacks)

    # create dataset
    dataset = CareamicsDataset(
        data_config=data_config,
        input_extractor=patch_extractor,
    )

    # extract tiles
    tiles: list[ImageRegionData] = []
    for i in range(n_tiles):
        tiles.append(dataset[i][0])

    # reshape array for testing
    arrays = [reshape_array(array[i], data_config.axes) for i in range(array.shape[0])]

    return np.stack(arrays, axis=0), tiles


@pytest.mark.parametrize(
    "axes, data_shape, expected_shape",
    [
        ("YX", (1, 1, 32, 64), (32, 64)),
        ("ZYX", (1, 1, 8, 32, 64), (8, 32, 64)),
        ("YXZ", (1, 1, 8, 32, 64), (8, 32, 64)),
        ("CYX", (1, 3, 32, 64), (3, 32, 64)),
        ("YXC", (1, 3, 32, 64), (3, 32, 64)),
        ("CZYX", (1, 3, 32, 64, 64), (3, 32, 64, 64)),
        ("ZCYX", (1, 3, 32, 64, 64), (3, 32, 64, 64)),
        ("ZYXC", (1, 3, 32, 64, 64), (3, 32, 64, 64)),
        ("SCYX", (8, 3, 32, 64), (8, 3, 32, 64)),
        ("SYXC", (8, 3, 32, 64), (8, 3, 32, 64)),
        ("SCZYX", (8, 3, 32, 64, 64), (8, 3, 32, 64, 64)),
        ("SZCYX", (8, 3, 32, 64, 64), (8, 3, 32, 64, 64)),
        ("ZSYXC", (8, 3, 32, 64, 64), (8, 3, 32, 64, 64)),
    ],
)
def test_update_data_shape(axes, data_shape, expected_shape):
    new_shape = _update_data_shape(axes, data_shape)
    assert new_shape == expected_shape


@pytest.mark.parametrize(
    "axes, data_shape, expected_chunks",
    [
        # axes are original data, can be STCZYX in any order
        # data_shape is in format SC(Z)YX with potential singleton dimensions
        # expected_chunks is in format SC(Z)YX as data is currently not reshaped
        # simple usual shapes
        ("YX", (1, 1, 32, 64), (32, 64)),
        ("YX", (1, 1, 128, 32), (128, 32)),
        ("ZYX", (1, 1, 32, 64, 64), (1, 64, 64)),
        ("ZYX", (1, 1, 64, 128, 64), (1, 128, 64)),
        ("CYX", (1, 5, 64, 64), (1, 64, 64)),
        ("SYX", (5, 1, 64, 256), (1, 64, 128)),
        ("SCYX", (8, 5, 64, 64), (1, 1, 64, 64)),
        ("SCZYX", (5, 5, 32, 256, 64), (1, 1, 1, 128, 64)),
        # different orders (but YX together)
        ("YXZ", (1, 1, 32, 64, 64), (1, 64, 64)),
        ("YXC", (1, 5, 64, 64), (1, 64, 64)),
        ("SYXZ", (1, 1, 32, 64, 64), (1, 1, 64, 64)),
        ("CSYX", (8, 5, 64, 64), (1, 1, 64, 64)),
        ("SZCYX", (8, 5, 512, 256, 64), (1, 1, 1, 128, 64)),
        # T dimension
        ("TYX", (5, 1, 64, 64), (1, 64, 64)),
        ("TCYX", (5, 3, 64, 64), (1, 1, 64, 64)),
        ("STYX", (5, 1, 64, 64), (1, 64, 64)),  # S and T together
        ("STCYX", (5, 3, 256, 64), (1, 1, 128, 64)),
    ],
)
def test_auto_chunks(axes, data_shape, expected_chunks):
    chunks = _auto_chunks(axes, data_shape)
    assert chunks == expected_chunks


@pytest.mark.parametrize("n_data", [1, 3])
@pytest.mark.parametrize(
    "axes, shape, shards, chunks, channels",
    [
        ("YX", (32, 32), (16, 16), (8, 8), None),
        ("CYX", (3, 32, 32), (1, 16, 16), (1, 8, 8), None),
        ("CYX", (3, 32, 32), (1, 16, 16), (1, 8, 8), [1]),
        ("CYX", (3, 32, 32), (1, 16, 16), (1, 8, 8), [0, 2]),
        ("ZYX", (16, 32, 32), (8, 16, 16), (4, 8, 8), None),
        ("CZYX", (3, 16, 32, 32), (1, 8, 16, 16), (1, 4, 8, 8), None),
        ("CZYX", (3, 16, 32, 32), (1, 8, 16, 16), (1, 4, 8, 8), [1]),
        ("CZYX", (3, 16, 32, 32), (1, 8, 16, 16), (1, 4, 8, 8), [0, 2]),
        ("SZYX", (5, 16, 32, 32), (1, 8, 16, 16), (1, 4, 8, 8), None),
        ("SCZYX", (5, 3, 16, 32, 32), (1, 1, 8, 16, 16), (1, 1, 4, 8, 8), None),
        ("SCZYX", (5, 3, 16, 32, 32), (1, 1, 8, 16, 16), (1, 1, 4, 8, 8), [1]),
        ("SCZYX", (5, 3, 16, 32, 32), (1, 1, 8, 16, 16), (1, 1, 4, 8, 8), [0, 2]),
    ],
)
def test_write_tile_identity(tmp_path, tiles, axes, shards, chunks, channels):
    """Test that `write_tile` correctly writes the data.

    No need to test with different axes order since the data coming to the writer
    is always in C(Z)YX format, with potential singleton dimensions.
    """
    arrays, tiles_list = tiles

    source_set = {tile.source for tile in tiles_list}

    # use writer to write predictions
    writer = WriteTilesZarr()
    for region in tiles_list:
        writer.write_tile(tmp_path, region)

    for src in source_set:
        filename = Path(src[len("file://") :]).parent.stem
        array_name = Path(src[len("file://") :]).name

        # check if zarr prediction exists
        zarr_path = tmp_path / f"{filename}_output.zarr"
        assert zarr_path.exists()

        # load array and compare with original
        g = zarr.open(zarr_path, mode="r")

        # check sharding and chunking
        if shards is not None:
            assert g[array_name].shards == shards
        if chunks is not None:
            assert g[array_name].chunks == chunks

        # pull array
        pred_array = g[array_name][:]
        data_idx = int(array_name.split("_")[-1])
        expected_array = arrays[data_idx]

        # if channels
        if channels is not None:
            expected_array = expected_array[:, channels]

        # zarr file writer does not save singleton dims if not present in original data
        if "C" not in axes:
            expected_array = expected_array.squeeze(axis=1)
        if "S" not in axes:
            expected_array = expected_array.squeeze(axis=0)

        np.testing.assert_allclose(pred_array, expected_array, rtol=1e-5, atol=0)


def test_write_from_array(tmp_path):
    """Test that writing from an array source raises an error."""
    arrays = [np.random.rand(32, 32).astype(np.float32) for _ in range(2)]
    image_stacks = load_arrays(
        source=arrays,
        axes="YX",
    )
    patch_extractor = PatchExtractor(image_stacks)

    strategy = TilingStrategy(
        data_shapes=[image_stacks[0].data_shape] * 2,
        patch_size=(8, 8),
        overlaps=(4, 4),
    )

    # use writer to write predictions
    writer = WriteTilesZarr()
    for region in gen_image_regions("YX", patch_extractor, strategy):
        writer.write_tile(tmp_path, region)

    # check if the stored zarr is close to the input
    if writer.current_store is not None:
        for key in writer.current_store.keys():
            np.testing.assert_allclose(
                arrays[int(key)], np.array(writer.current_store[key])
            )
    else:
        raise ValueError("The Zarr tile writer store is None!")


def test_write_from_tiff(tmp_path):
    """Test that writing from a tiff source raises an error."""
    # save tiffs
    arrays = [np.random.rand(32, 32).astype(np.float32) for _ in range(2)]
    files = [tmp_path / f"test_{i}.tiff" for i in range(2)]
    for file, array in zip(files, arrays, strict=True):
        tifffile.imwrite(file, array)

    image_stacks = load_tiffs(
        source=files,
        axes="YX",
    )
    patch_extractor = PatchExtractor(image_stacks)

    strategy = TilingStrategy(
        data_shapes=[image_stacks[0].data_shape] * 2,
        patch_size=(8, 8),
        overlaps=(4, 4),
    )

    # use writer to write predictions
    writer = WriteTilesZarr()
    for region in gen_image_regions("YX", patch_extractor, strategy):
        writer.write_tile(tmp_path, region)

    # check if the stored zarr is close to the input
    if writer.current_store is not None:
        for key in writer.current_store.keys():
            np.testing.assert_allclose(
                arrays[int(key)], np.array(writer.current_store[key])
            )
    else:
        raise ValueError("The Zarr tile writer store is None!")



"""Convenience function to create N2V configurations."""

from collections.abc import Sequence
from typing import Any, Literal

from careamics.config.ng_configs import N2VConfiguration
from careamics.config.support import (
    SupportedPixelManipulation,
    SupportedTransform,
)
from careamics.config.transformations import (
    N2VManipulateConfig,
    XYFlipConfig,
    XYRandomRotate90Config,
)

from .algorithm_factory import create_algorithm_configuration
from .data_factory import create_ng_data_configuration, list_spatial_augmentations
from .training_factory import create_training_configuration, update_trainer_params


def create_n2v_config(
    # mandatory parameters
    experiment_name: str,
    data_type: Literal["array", "tiff", "zarr", "czi", "custom"],
    axes: str,
    patch_size: Sequence[int],
    batch_size: int,
    # optional parameters
    num_epochs: int = 30,  # not too high, in case data is very large
    num_steps: int | None = None,
    augmentations: Sequence[Literal["x_flip", "y_flip", "rotate_90"]] | None = None,
    use_n2v2: bool = False,
    n_channels: int | None = None,
) -> N2VConfiguration:
    """
    Create a configuration for training N2V.

    To activate N2V2, set `use_n2v2` to True.

    The `axes` parameters must reflect the actual axes and axis order from the data,
    and should be the same throughout all images. The accepted axes are STCZYX. If "C"
    is in `axes`, then you need to set `n_channels` to the number of channels.

    By default, CAREamics will go through the entire training data once per epoch. For
    large datasets, this can lead to very long epochs. To limit the number of batches
    per epoch, set the `num_steps` parameter to the desired number of batches.

    If the content of your data is expected to always have the same orientation,
    consider disabling certain augmentations. By default `augmentations=None` will apply
    random flips along X and Y, and random 90 degrees rotations in the XY plane. To
    disable augmentations, set `augmentations=[]`.

    See `create_advanced_n2v_config` for more parameters.

    Parameters
    ----------
    experiment_name : str
        Name of the experiment.
    data_type : Literal["array", "tiff", "zarr", "czi", "custom"]
        Type of the data.
    axes : str
        Axes of the data (e.g. SYX).
    patch_size : Sequence[int]
        Size of the patches along the spatial dimensions (e.g. [64, 64]).
    batch_size : int
        Batch size.
    num_epochs : int, default=30
        Number of epochs to train for.
    num_steps : int, default=None
        Number of batches in 1 epoch.
    augmentations : Sequence of {"x_flip", "y_flip", "rotate_90"}, default=None
        List of augmentations to apply. If `None`, all augmentations are applied.
    use_n2v2 : bool, default=False
        Whether to use N2V2.
    n_channels : int or None, default=None
        Number of channels (in and out).

    Returns
    -------
    N2VConfiguration
        Configuration for training N2V.
    """
    return create_advanced_n2v_config(
        experiment_name=experiment_name,
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        num_epochs=num_epochs,
        num_steps=num_steps,
        augmentations=augmentations,
        use_n2v2=use_n2v2,
        n_channels=n_channels,
    )


def create_structn2v_config(
    # mandatory parameters
    experiment_name: str,
    data_type: Literal["array", "tiff", "zarr", "czi", "custom"],
    axes: str,
    patch_size: Sequence[int],
    batch_size: int,
    # struct n2v
    struct_n2v_axis: Literal["horizontal", "vertical"],
    struct_n2v_span: int = 5,
    # optional parameters
    num_epochs: int = 30,
    num_steps: int | None = None,
    # TODO no rotation until we support 2D masks for structN2V
    use_n2v2: bool = False,
    n_channels: int | None = None,
) -> N2VConfiguration:
    """
    Create a configuration for training structN2V.

    The structN2V mask is applied a horizontal or vertical axis, with extent defined by
    `struct_n2v_span` (default=5, leading to a mask of size 11). For structN2V,
    augmentations are disabled.

    To activate N2V2, set `use_n2v2` to True.

    The `axes` parameters must reflect the actual axes and axis order from the data,
    and should be the same throughout all images. The accepted axes are STCZYX. If "C"
    is in `axes`, then you need to set `n_channels` to the number of channels.

    `patch_size` is only along the spatial dimensions and should be of length 3 if "Z"
    is present in `axes`, otherwise of length 2.

    By default, CAREamics will go through the entire training data once per epoch. For
    large datasets, this can lead to very long epochs. To limit the number of batches
    per epoch, set the `num_steps` parameter to the desired number of batches.

    See `create_advanced_n2v_config` for more parameters.

    Parameters
    ----------
    experiment_name : str
        Name of the experiment.
    data_type : Literal["array", "tiff", "zarr", "czi", "custom"]
        Type of the data.
    axes : str
        Axes of the data (e.g. SYX).
    patch_size : Sequence[int]
        Size of the patches along the spatial dimensions (e.g. [64, 64]).
    batch_size : int
        Batch size.
    struct_n2v_axis : Literal["horizontal", "vertical"]
        Axis along which to apply structN2V mask.
    struct_n2v_span : int, default=5
        Span of the structN2V mask.
    num_epochs : int, default=30
        Number of epochs to train for.
    num_steps : int, default=None
        Number of batches in 1 epoch.
    use_n2v2 : bool, default=False
        Whether to use N2V2.
    n_channels : int or None, default=None
        Number of channels (in and out).

    Returns
    -------
    N2VConfiguration
        Configuration for training structN2V.
    """
    return create_advanced_n2v_config(
        experiment_name=experiment_name,
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        num_epochs=num_epochs,
        num_steps=num_steps,
        augmentations=[],
        use_n2v2=use_n2v2,
        n_channels=n_channels,
        struct_n2v_axis=struct_n2v_axis,
        struct_n2v_span=struct_n2v_span,
    )


# TODO reorganize docstring parameters
def create_advanced_n2v_config(
    experiment_name: str,
    data_type: Literal["array", "tiff", "zarr", "czi", "custom"],
    axes: str,
    patch_size: Sequence[int],
    batch_size: int,
    # optional parameters
    num_epochs: int = 30,
    num_steps: int | None = None,
    n_channels: int | None = None,
    augmentations: Sequence[Literal["x_flip", "y_flip", "rotate_90"]] | None = None,
    # advanced parameters
    in_memory: bool | None = None,
    channels: Sequence[int] | None = None,
    independent_channels: bool = True,
    normalization: Literal["mean_std", "minmax", "quantile", "none"] = "mean_std",
    normalization_params: dict[str, Any] | None = None,
    # - N2V specific
    use_n2v2: bool = False,
    roi_size: int = 11,
    masked_pixel_percentage: float = 0.2,
    # - structN2V specific
    struct_n2v_axis: Literal["horizontal", "vertical", "none"] = "none",
    struct_n2v_span: int = 5,
    # - Lightning parameters
    num_workers: int = 0,
    trainer_params: dict | None = None,
    model_params: dict | None = None,
    optimizer: Literal["Adam", "Adamax", "SGD"] = "Adam",
    optimizer_params: dict[str, Any] | None = None,
    lr_scheduler: Literal["ReduceLROnPlateau", "StepLR"] = "ReduceLROnPlateau",
    lr_scheduler_params: dict[str, Any] | None = None,
    train_dataloader_params: dict[str, Any] | None = None,
    val_dataloader_params: dict[str, Any] | None = None,
    checkpoint_params: dict[str, Any] | None = None,
    logger: Literal["wandb", "tensorboard", "none"] = "none",
    # - reproducibility
    seed: int | None = None,
) -> N2VConfiguration:
    """
    Create a configuration for training Noise2Void.

    N2V uses a UNet model to denoise images in a self-supervised manner. To use its
    variants structN2V and N2V2, set the `struct_n2v_axis` and `struct_n2v_span`
    (structN2V) parameters, or set `use_n2v2` to True (N2V2).

    N2V2 modifies the UNet architecture by adding blur pool layers and removes the skip
    connections, thus removing checkboard artefacts. StructN2V is used when vertical
    or horizontal correlations are present in the noise; it applies an additional mask
    to the manipulated pixel neighbors.

    If "Z" is present in `axes`, then `patch_size` must be a list of length 3, otherwise
    2.

    If "C" is present in `axes`, then you need to set `n_channels` to the number of
    channels.

    By default, all channels are trained independently. To train all channels together,
    set `independent_channels` to False.

    By default, the augmentations applied are random flips along X or Y, and random
    90 degrees rotations in the XY plane. To disable the augmentations, simply pass an
    empty list.

    The `roi_size` parameter specifies the size of the area around each pixel that will
    be manipulated by N2V. The `masked_pixel_percentage` parameter specifies how many
    pixels per patch will be manipulated.

    If you pass "horizontal" or "vertical" to `struct_n2v_axis`, then structN2V mask
    will be applied to each manipulated pixel.

    The parameters of the UNet can be specified in the `model_params` (passed as a
    parameter-value dictionary). Note that `use_n2v2` and 'n_channels' override the
    corresponding parameters passed in `model_params`.

    Note that `num_workers` is applied to all dataloaders unless explicitly overridden
    in the respective dataloader parameters.

    Parameters
    ----------
    experiment_name : str
        Name of the experiment.
    data_type : Literal["array", "tiff", "zarr", "czi", "custom"]
        Type of the data.
    axes : str
        Axes of the data (e.g. SYX).
    patch_size : Sequence[int]
        Size of the patches along the spatial dimensions (e.g. [64, 64]).
    batch_size : int
        Batch size.
    num_epochs : int, default=30
        Number of epochs to train for. If provided, this will be added to
        trainer_params.
    num_steps : int | None, default=None
        Number of batches in 1 epoch. If provided, this will be added to trainer_params.
        Translates to `limit_train_batches` in PyTorch Lightning Trainer. See relevant
        documentation for more details.
    n_channels : int | None, default=None
        Number of channels (in and out). If `channels` is specified, then the number of
        channels is inferred from its length and this parameter is ignored.
    augmentations : Sequence[{"x_flip", "y_flip", "rotate_90"}] | None, default=None
        List of transforms to apply, either both or one of XYFlipConfig and
        XYRandomRotate90Config. By default, it applies both XYFlip (on X and Y)
        and XYRandomRotate90 (in XY) to the images.
    in_memory : bool | None, default=None
        Whether to load all data into memory. This is only supported for 'array',
        'tiff' and 'custom' data types. If `None`, defaults to `True` for 'array',
        'tiff' and `custom`, and `False` for 'zarr' and 'czi' data types. Must be `True`
        for `array`.
    channels : Sequence[int] | None, default=None
        List of channels to use. If `None`, all channels are used.
    independent_channels : bool, default=True
        Whether to train all channels independently.
    normalization : {"mean_std", "minmax", "quantile", "none"}, default="mean_std"
        Normalization strategy to use.
    normalization_params : dict[str, Any] | None, default=None
        Strategy-specific normalization parameters. If None, default values are used.
        For "mean_std": {"input_means": [...], "input_stds": [...]} (optional)
        For "minmax": {"input_mins": [...], "input_maxes": [...]} (optional)
        For "quantile": {"lower_quantile": 0.01, "upper_quantile": 0.99} (optional)
        For "none": No parameters needed.
    use_n2v2 : bool, default=False
        Whether to use N2V2.
    roi_size : int, default=11
        N2V pixel manipulation area.
    masked_pixel_percentage : float, default=0.2
        Percentage of pixels masked in each patch.
    struct_n2v_axis : Literal["horizontal", "vertical", "none"], default="none"
        Axis along which to apply structN2V mask.
    struct_n2v_span : int, default=5
        Span of the structN2V mask.
    num_workers : int, default=0
        Number of workers for data loading. Unless explicitly overridden in
        `train_dataloader_params` and `val_dataloader_params`, this will be applied to
        all dataloaders.
    trainer_params : dict | None, default=None
        Parameters for the trainer, see the relevant documentation.
    model_params : dict | None, default=None
        UNetModel parameters.
    optimizer : Literal["Adam", "Adamax", "SGD"], default="Adam"
        Optimizer to use.
    optimizer_params : dict[str, Any] | None, default=None
        Parameters for the optimizer, see PyTorch documentation for more details.
    lr_scheduler : Literal["ReduceLROnPlateau", "StepLR"], default="ReduceLROnPlateau"
        Learning rate scheduler to use.
    lr_scheduler_params : dict[str, Any] | None, default=None
        Parameters for the learning rate scheduler, see PyTorch documentation for more
        details.
    train_dataloader_params : dict[str, Any] | None, default=None
        Parameters for the training dataloader, see the PyTorch docs for `DataLoader`.
        If left as `None`, `{"shuffle": True}` will be used.
    val_dataloader_params : dict[str, Any] | None, default=None
        Parameters for the validation dataloader, see PyTorch the docs for `DataLoader`.
    checkpoint_params : dict[str, Any] | None, default=None
        Parameters for the checkpoint callback, see PyTorch Lightning documentation
        (`ModelCheckpoint`) for the list of available parameters.
    logger : Literal["wandb", "tensorboard", "none"], default="none"
        Logger to use.
    seed : int | None, default=None
        Random seed for reproducibility.

    Returns
    -------
    N2VConfiguration
        Configuration for training N2V.
    """
    # if there are channels, we need to specify their number
    channels_present = "C" in axes

    if channels_present and (n_channels is None and channels is None):
        raise ValueError(
            "`n_channels` or `channels` must be specified when using channels."
        )
    elif not channels_present and (n_channels is not None and n_channels > 1):
        raise ValueError(
            f"C is not present in the axes, but number of channels is specified "
            f"(got {n_channels} channel)."
        )

    if n_channels is not None and channels is not None:
        if n_channels != len(channels):
            raise ValueError(
                f"Number of channels ({n_channels}) does not match length of "
                f"`channels` ({len(channels)}). Only specify `channels`."
            )

    if n_channels is None:
        n_channels = 1 if channels is None else len(channels)

    # normalization
    norm_config = {"name": normalization}
    if normalization_params is not None:
        norm_config.update(normalization_params)

    # augmentations
    augs: list[XYFlipConfig | XYRandomRotate90Config] | None = None
    if augmentations is not None:
        augs = []

        x_flip_present = "x_flip" in augmentations
        y_flip_present = "y_flip" in augmentations
        rotate_90_present = "rotate_90" in augmentations

        if x_flip_present or y_flip_present:
            augs.append(
                XYFlipConfig(
                    flip_x=x_flip_present,
                    flip_y=y_flip_present,
                    seed=seed,
                )
            )
        if rotate_90_present:
            augs.append(XYRandomRotate90Config(seed=seed))
    spatial_transforms = list_spatial_augmentations(augs)

    # data
    data_config = create_ng_data_configuration(
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        augmentations=spatial_transforms,
        normalization=norm_config,
        channels=channels,
        in_memory=in_memory,
        num_workers=num_workers,
        train_dataloader_params=train_dataloader_params,
        val_dataloader_params=val_dataloader_params,
        seed=seed,
    )

    # algorithm
    algorithm_params = create_algorithm_configuration(
        dimensions=3 if data_config.is_3D() else 2,
        algorithm="n2v",
        loss="n2v",
        independent_channels=independent_channels,
        n_channels_in=n_channels,
        n_channels_out=n_channels,
        use_n2v2=use_n2v2,
        model_params=model_params,
        optimizer=optimizer,
        optimizer_params=optimizer_params,
        lr_scheduler=lr_scheduler,
        lr_scheduler_params=lr_scheduler_params,
    )

    # create the N2VManipulate transform using the supplied parameters
    n2v_transform = N2VManipulateConfig(  # TODO should be seeded
        name=SupportedTransform.N2V_MANIPULATE.value,
        strategy=(
            SupportedPixelManipulation.MEDIAN.value
            if use_n2v2
            else SupportedPixelManipulation.UNIFORM.value
        ),
        roi_size=roi_size,
        masked_pixel_percentage=masked_pixel_percentage,
        struct_mask_axis=struct_n2v_axis,
        struct_mask_span=struct_n2v_span,
    )
    algorithm_params["n2v_config"] = n2v_transform

    # training
    final_trainer_params = update_trainer_params(
        trainer_params=trainer_params,
        num_epochs=num_epochs,
        num_steps=num_steps,
    )
    training_params = create_training_configuration(
        trainer_params=final_trainer_params,
        logger=logger,
        checkpoint_params=checkpoint_params,
    )

    return N2VConfiguration(
        experiment_name=experiment_name,
        algorithm_config=algorithm_params,
        data_config=data_config,
        training_config=training_params,
    )



"""
Loss submodule.

This submodule contains the various losses used in CAREamics.
"""

import torch
from torch.nn import L1Loss, MSELoss

from careamics.models.lvae.noise_models import GaussianMixtureNoiseModel


def mse_loss(source: torch.Tensor, target: torch.Tensor, *args) -> torch.Tensor:
    """
    Mean squared error loss.

    Parameters
    ----------
    source : torch.Tensor
        Source patches.
    target : torch.Tensor
        Target patches.
    *args : Any
        Additional arguments.

    Returns
    -------
    torch.Tensor
        Loss value.
    """
    loss = MSELoss()
    return loss(source, target)


def n2v_loss(
    manipulated_batch: torch.Tensor,
    original_batch: torch.Tensor,
    masks: torch.Tensor,
    *args,
) -> torch.Tensor:
    """
    N2V Loss function described in A Krull et al 2018.

    Parameters
    ----------
    manipulated_batch : torch.Tensor
        Batch after manipulation function applied.
    original_batch : torch.Tensor
        Original images.
    masks : torch.Tensor
        Coordinates of changed pixels.
    *args : Any
        Additional arguments.

    Returns
    -------
    torch.Tensor
        Loss value.
    """
    errors = (original_batch - manipulated_batch) ** 2
    # Average over pixels and batch
    loss = torch.sum(errors * masks) / torch.sum(masks)
    return loss  # TODO change output to dict ?


def pn2v_loss(
    samples: torch.Tensor,
    labels: torch.Tensor,
    masks: torch.Tensor,
    noise_model: GaussianMixtureNoiseModel,
) -> torch.Tensor:
    """
    Probabilistic N2V loss function described in A Krull et al., CVF (2019).

    Parameters
    ----------
    samples : torch.Tensor # TODO this naming is confusing
        Predicted pixel values from the network.
    labels : torch.Tensor
        Original pixel values.
    masks : torch.Tensor
        Coordinates of manipulated pixels.
    noise_model : GaussianMixtureNoiseModel
        Noise model for computing likelihood.

    Returns
    -------
    torch.Tensor
        Loss value.
    """
    likelihoods = noise_model.likelihood(labels, samples)
    likelihoods_avg = torch.log(torch.mean(likelihoods, dim=1, keepdim=True))

    # Average over pixels and batch
    loss = -torch.sum(likelihoods_avg * masks) / torch.sum(masks)
    return loss


def mae_loss(samples: torch.Tensor, labels: torch.Tensor, *args) -> torch.Tensor:
    """
    N2N Loss function described in to J Lehtinen et al 2018.

    Parameters
    ----------
    samples : torch.Tensor
        Raw patches.
    labels : torch.Tensor
        Different subset of noisy patches.
    *args : Any
        Additional arguments.

    Returns
    -------
    torch.Tensor
        Loss value.
    """
    loss = L1Loss()
    return loss(samples, labels)


# def dice_loss(
#     samples: torch.Tensor, labels: torch.Tensor, mode: str = "multiclass"
# ) -> torch.Tensor:
#     """Dice loss function."""
#     return DiceLoss(mode=mode)(samples, labels.long())



"""Test `WriteImage` class."""

from pathlib import Path

import numpy as np
import pytest
from numpy.typing import NDArray

from careamics.dataset_ng.dataset import ImageRegionData
from careamics.lightning.dataset_ng.callbacks.prediction_writer import (
    WriteImage,
    create_write_file_path,
)

# TODO test when multiple sample


@pytest.fixture
def write_func():
    def a_write_func(file_path: Path, img: NDArray, *args, **kwargs) -> None:
        """Do-nothing write function."""
        return

    return a_write_func


@pytest.fixture
def write_image_strategy(write_func) -> WriteImage:
    """
    Initialized `WriteImage` class.

    Parameters
    ----------
    write_func : WriteFunc
        Write function. (Comes from fixture).

    Returns
    -------
    CacheTiles
        Initialized `CacheTiles` class.
    """
    write_extension = ".ext"
    write_func_kwargs = {}
    return WriteImage(
        write_func=write_func,
        write_extension=write_extension,
        write_func_kwargs=write_func_kwargs,
    )


def test_cache_tiles_init(write_func, write_image_strategy):
    """
    Test `WriteImage` initializes as expected.
    """
    assert write_image_strategy.write_func is write_func
    assert write_image_strategy.write_extension == ".ext"
    assert write_image_strategy.write_func_kwargs == {}


def test_write_image_batch(write_image_strategy, ordered_array, mocker):
    """Test writing a batch."""
    array = ordered_array((5, 8, 8))

    prediction = [
        ImageRegionData(
            source="array.tiff",
            data=array[i : i + 1],  # keep S dim as singleton to mock C dim
            data_shape=array.shape,
            dtype=np.float32,
            axes="SYX",
            region_spec={
                "data_idx": 0,
                "sample_idx": i,
                "coords": (0, 0),
                "patch_size": (8, 8),
            },
            additional_metadata={},
        )
        for i in range(array.shape[0])
    ]

    # spy on write function
    spy = mocker.spy(write_image_strategy, "write_func")

    # call write batch
    dirpath = Path("out_dir")
    write_image_strategy.write_batch(
        dirpath=dirpath,
        predictions=prediction,
    )

    # check call and arguments
    spy.assert_called_once()

    call_args = spy.call_args.kwargs
    expected_file_path = create_write_file_path(
        dirpath=dirpath,
        file_path=prediction[0].source,
        write_extension=write_image_strategy.write_extension,
        postfix="_0",
    )
    assert call_args["file_path"] == expected_file_path
    np.testing.assert_array_equal(call_args["img"], array)



import pytest

from careamics.config.validators import check_axes_validity, check_czi_axes_validity


@pytest.mark.parametrize(
    "axes, valid",
    [
        # Passing
        ("yx", True),
        ("Yx", True),
        ("Zyx", True),
        ("STYX", True),
        ("CYX", True),
        ("YXC", True),
        ("TzYX", True),
        ("SZYX", True),
        ("STZYX", True),
        ("XY", True),
        ("YXT", True),
        ("ZTYX", True),
        # non consecutive XY
        ("YZX", False),
        ("YZCXT", False),
        # too few axes
        ("", False),
        ("X", False),
        # no yx axes
        ("ZT", False),
        ("ZY", False),
        # repeating characters
        ("YYX", False),
        ("YXY", False),
        # invalid characters
        ("YXm", False),
        ("1YX", False),
    ],
)
def test_check_axes_validity(axes, valid):
    """Test if axes are valid"""
    if valid:
        check_axes_validity(axes)
    else:
        with pytest.raises((ValueError, NotImplementedError)):
            check_axes_validity(axes)


@pytest.mark.parametrize(
    "axes, expected",
    [
        ("SCZYX", True),
        ("SCTYX", True),
        ("SCYX", True),
        ("CTYX", False),  # missing S axis
        ("SCZ", False),  # missing YX axes
        ("SCZYXT", False),  # extra axis
        ("TCYX", False),  # wrong order
    ],
)
def test_check_czi_axes_validity(axes: str, expected: bool):
    """Test `are_axes_valid` function."""
    assert check_czi_axes_validity(axes) == expected



"""Zarr-related fixtures"""

from pathlib import Path
from typing import TypedDict

import numpy as np
import pytest
import zarr
from numpy.typing import NDArray

# TODO create an OME-Zarr fixture


class ZarrSource(TypedDict):
    zarr_file: str | Path
    array_path: str


@pytest.fixture
def ome_zarr_url() -> str:
    """URL to a public OME-Zarr for testing."""
    return "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0062A/6001240.zarr"


@pytest.fixture
def arrays() -> NDArray:
    return np.arange(3 * 16 * 16).reshape((3, 16, 16))


@pytest.fixture
def zarr_linear(tmp_path, arrays) -> list[ZarrSource]:
    """Three arrays in a single zarr group at the root."""
    path = tmp_path / "linear.zarr"
    zarr_file = zarr.create_group(str(path.absolute()))

    # write arrays to zarr
    source = []
    for i in range(arrays.shape[0]):
        zarr_file.create_array(f"array{i}", data=arrays[i], chunks=(8, 8))

        source.append(
            {
                "zarr_file": str(path.absolute()),
                "array_path": f"array{i}",
            }
        )

    return source


@pytest.fixture
def zarr_groups(tmp_path, arrays) -> list[ZarrSource]:
    """Two groups in a single zarr file, each containing arrays."""
    path = tmp_path / "groups.zarr"
    zarr_file = zarr.create_group(path)

    # write arrays to zarr in different groups
    source = []
    for i in range(arrays.shape[0]):
        if i % 2 == 0:
            group_name = "groupA"
        else:
            group_name = "groupB"

        if group_name not in zarr_file:
            group = zarr_file.create_group(group_name)
        else:
            group = zarr_file[group_name]

        group.create_array(f"array{i}", data=arrays[i], chunks=(8, 8))

        source.append(
            {
                "zarr_file": str(path.absolute()),
                "array_path": f"{group_name}/array{i}",
            }
        )

    return source


@pytest.fixture
def zarr_multiple(tmp_path, arrays) -> list[ZarrSource]:
    """Arrays in the root of two different zarr files."""
    source = []

    # write each array to its own zarr file
    for i in range(arrays.shape[0]):

        if i % 2 == 0:
            path = tmp_path / "even_arrays.zarr"
        else:
            path = tmp_path / "odd_arrays.zarr"

        if not path.exists():
            zarr_file = zarr.create_group(path)
        else:
            zarr_file = zarr.open(path)

        zarr_file.create_array(f"array{i}", data=arrays[i], chunks=(8, 8))
        source.append(
            {
                "zarr_file": str(path.absolute()),
                "array_path": f"array{i}",
            }
        )

    return source



"""PN2V Algorithm configuration."""

from typing import Annotated, Literal, Self

from bioimageio.spec.generic.v0_3 import CiteEntry
from pydantic import AfterValidator, ConfigDict, model_validator

from careamics.config.architectures import UNetConfig
from careamics.config.noise_model import GaussianMixtureNMConfig
from careamics.config.support import SupportedPixelManipulation, SupportedStructAxis
from careamics.config.transformations import N2VManipulateConfig
from careamics.config.validators import (
    model_without_final_activation,
)

from .unet_algorithm_config import UNetBasedAlgorithm

PN2V = "Probabilistic Noise2Void"
PN2V2 = "PN2V2"
STRUCT_PN2V = "StructPN2V"
STRUCT_PN2V2 = "StructPN2V2"

PN2V_REF = CiteEntry(
    text="Krull, A., Vicar, T., Prakash, M., Lalit, M. and Jug, F., 2020. "
    '"Probabilistic noise2void: Unsupervised content-aware denoising". '
    "Frontiers in Computer Science, 2, p.5.",
    doi="10.3389/fcomp.2020.00005",
)

N2V2_REF = CiteEntry(
    text="Höck, E., Buchholz, T.O., Brachmann, A., Jug, F. and Freytag, A., "
    '2022. "N2V2 - Fixing Noise2Void checkerboard artifacts with modified '
    'sampling strategies and a tweaked network architecture". In European '
    "Conference on Computer Vision (pp. 503-518).",
    doi="10.1007/978-3-031-25069-9_33",
)

STRUCTN2V_REF = CiteEntry(
    text="Broaddus, C., Krull, A., Weigert, M., Schmidt, U. and Myers, G., 2020."
    '"Removing structured noise with self-supervised blind-spot '
    'networks". In 2020 IEEE 17th International Symposium on Biomedical '
    "Imaging (ISBI) (pp. 159-163).",
    doi="10.1109/isbi45749.2020.9098336",
)

PN2V_DESCRIPTION = (
    "Probabilistic Noise2Void (PN2V) is a UNet-based self-supervised algorithm that "
    "extends Noise2Void by incorporating a probabilistic noise model to estimate the "
    "posterior distribution of each pixel more precisely. Like N2V, it uses blind-spot "
    "training where random pixels are selected in patches during training and their "
    "value replaced by a neighboring pixel value. The model is then trained to predict "
    "the original pixel value. PN2V additionally uses a noise model to better capture "
    "the noise characteristics and provide more accurate denoising by modeling the "
    "likelihood of pixel intensities."
)

PN2V2_DESCRIPTION = (
    "PN2V2 is a variant of Probabilistic Noise2Void. "
    + PN2V_DESCRIPTION
    + "\nPN2V2 introduces blur-pool layers and removed skip "
    "connections in the UNet architecture to remove checkboard "
    "artefacts, a common artefacts occurring in Noise2Void variants."
)

STRUCT_PN2V_DESCRIPTION = (
    "StructPN2V is a variant of Probabilistic Noise2Void. "
    + PN2V_DESCRIPTION
    + "\nStructPN2V uses a linear mask (horizontal or vertical) to replace "
    "the pixel values of neighbors of the masked pixels by a random "
    "value. Such masking allows removing 1D structured noise from the "
    "the images, the main failure case of the original N2V variants."
)

STRUCT_PN2V2_DESCRIPTION = (
    "StructPN2V2 is a variant of Probabilistic Noise2Void that uses both "
    "structN2V and N2V2. "
    + PN2V_DESCRIPTION
    + "\nStructPN2V2 uses a linear mask (horizontal or vertical) to replace "
    "the pixel values of neighbors of the masked pixels by a random "
    "value. Such masking allows removing 1D structured noise from the "
    "the images, the main failure case of the original N2V variants."
    "\nPN2V2 introduces blur-pool layers and removed skip connections in "
    "the UNet architecture to remove checkboard artefacts, a common "
    "artefacts occurring in Noise2Void variants."
)


class PN2VAlgorithm(UNetBasedAlgorithm):
    """PN2V Algorithm configuration."""

    model_config = ConfigDict(validate_assignment=True)

    algorithm: Literal["pn2v"] = "pn2v"
    """PN2V Algorithm name."""

    loss: Literal["pn2v"] = "pn2v"
    """PN2V loss function (uses N2V loss with noise model)."""

    n2v_config: N2VManipulateConfig = N2VManipulateConfig()

    noise_model: GaussianMixtureNMConfig
    """Noise model configuration for probabilistic denoising."""

    model: Annotated[
        UNetConfig,
        # AfterValidator(model_matching_in_out_channels),
        # TODO for pn2v channel handling needs to be changed
        AfterValidator(model_without_final_activation),
    ]

    @model_validator(mode="after")
    def validate_n2v2(self) -> Self:
        """Validate that the N2V2 strategy and models are set correctly.

        Returns
        -------
        Self
            The validated configuration.

        Raises
        ------
        ValueError
            If N2V2 is used with the wrong pixel manipulation strategy.
        """
        if self.model.n2v2:
            if self.n2v_config.strategy != SupportedPixelManipulation.MEDIAN.value:
                raise ValueError(
                    f"PN2V2 can only be used with the "
                    f"{SupportedPixelManipulation.MEDIAN} pixel manipulation strategy. "
                    f"Change the `strategy` parameters in `n2v_config` to "
                    f"{SupportedPixelManipulation.MEDIAN}."
                )
        else:
            if self.n2v_config.strategy != SupportedPixelManipulation.UNIFORM.value:
                raise ValueError(
                    f"PN2V can only be used with the "
                    f"{SupportedPixelManipulation.UNIFORM} pixel manipulation strategy."
                    f" Change the `strategy` parameters in `n2v_config` to "
                    f"{SupportedPixelManipulation.UNIFORM}."
                )
        return self

    def set_n2v2(self, use_n2v2: bool) -> None:
        """
        Set the configuration to use PN2V2 or the vanilla Probabilistic Noise2Void.

        This method ensures that PN2V2 is set correctly and remain coherent, as opposed
        to setting the different parameters individually.

        Parameters
        ----------
        use_n2v2 : bool
            Whether to use PN2V2.
        """
        if use_n2v2:
            self.n2v_config.strategy = SupportedPixelManipulation.MEDIAN.value
            self.model.n2v2 = True
        else:
            self.n2v_config.strategy = SupportedPixelManipulation.UNIFORM.value
            self.model.n2v2 = False

    def is_struct_n2v(self) -> bool:
        """Check if the configuration is using structPN2V.

        Returns
        -------
        bool
            Whether the configuration is using structPN2V.
        """
        return self.n2v_config.struct_mask_axis != SupportedStructAxis.NONE.value

    def get_algorithm_friendly_name(self) -> str:
        """
        Get the friendly name of the algorithm.

        Returns
        -------
        str
            Friendly name.
        """
        use_n2v2 = self.model.n2v2
        use_structN2V = self.is_struct_n2v()

        if use_n2v2 and use_structN2V:
            return STRUCT_PN2V2
        elif use_n2v2:
            return PN2V2
        elif use_structN2V:
            return STRUCT_PN2V
        else:
            return PN2V

    def get_algorithm_keywords(self) -> list[str]:
        """
        Get algorithm keywords.

        Returns
        -------
        list[str]
            List of keywords.
        """
        use_n2v2 = self.model.n2v2
        use_structN2V = self.is_struct_n2v()

        keywords = [
            "denoising",
            "restoration",
            "UNet",
            "3D" if self.model.is_3D() else "2D",
            "CAREamics",
            "pytorch",
            PN2V,
            "probabilistic",
            "noise model",
        ]

        if use_n2v2:
            keywords.append(PN2V2)
        if use_structN2V:
            keywords.append(STRUCT_PN2V)

        return keywords

    def get_algorithm_references(self) -> str:
        """
        Get the algorithm references.

        This is used to generate the README of the BioImage Model Zoo export.

        Returns
        -------
        str
            Algorithm references.
        """
        use_n2v2 = self.model.n2v2
        use_structN2V = self.is_struct_n2v()

        references = [
            PN2V_REF.text + " doi: " + PN2V_REF.doi,
            N2V2_REF.text + " doi: " + N2V2_REF.doi,
            STRUCTN2V_REF.text + " doi: " + STRUCTN2V_REF.doi,
        ]

        # return the (struct)PN2V(2) references
        if use_n2v2 and use_structN2V:
            return "\n".join(references)
        elif use_n2v2:
            references.pop(-1)
            return "\n".join(references)
        elif use_structN2V:
            references.pop(-2)
            return "\n".join(references)
        else:
            return references[0]

    def get_algorithm_citations(self) -> list[CiteEntry]:
        """
        Return a list of citation entries of the current algorithm.

        This is used to generate the model description for the BioImage Model Zoo.

        Returns
        -------
        List[CiteEntry]
            List of citation entries.
        """
        use_n2v2 = self.model.n2v2
        use_structN2V = self.is_struct_n2v()

        references = [PN2V_REF]

        if use_n2v2:
            references.append(N2V2_REF)

        if use_structN2V:
            references.append(STRUCTN2V_REF)

        return references

    def get_algorithm_description(self) -> str:
        """
        Return a description of the algorithm.

        This method is used to generate the README of the BioImage Model Zoo export.

        Returns
        -------
        str
            Description of the algorithm.
        """
        use_n2v2 = self.model.n2v2
        use_structN2V = self.is_struct_n2v()

        if use_n2v2 and use_structN2V:
            return STRUCT_PN2V2_DESCRIPTION
        elif use_n2v2:
            return PN2V2_DESCRIPTION
        elif use_structN2V:
            return STRUCT_PN2V_DESCRIPTION
        else:
            return PN2V_DESCRIPTION
